"""
Module for function spaces of generalized Jacobi type

"""

import numpy as np
import sympy as sp
from scipy.special import eval_jacobi, roots_jacobi #, gamma
from shenfun.config import config
from shenfun.spectralbase import SpectralBase, getCompositeBase, getBCGeneric, \
    BoundaryConditions
from shenfun.matrixbase import SparseMatrix
from .recursions import h, n

xp = sp.Symbol('x', real=True)
m, n, k = sp.symbols('m,n,k', real=True, integer=True)

#pylint: disable=method-hidden,no-else-return,not-callable,abstract-method,no-member,cyclic-import

bases = ['Orthogonal',
         'CompactDirichlet',
         'CompactNeumann',
         'UpperDirichlet',
         'LowerDirichlet',
         'Generic']
bcbases = ['BCGeneric']
testbases = ['Phi1', 'Phi2', 'Phi3', 'Phi4']
__all__ = bases + bcbases + testbases + ['JacobiBase']


class JacobiBase(SpectralBase):
    r"""Abstract base class for Jacobi function spaces
    """
    def __init__(self, N, quad='', alpha=0, beta=0, padding_factor=1, domain=(-1., 1.), dtype=None,
                 dealias_direct=False, coordinates=None):
        SpectralBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype,
                              padding_factor=padding_factor, dealias_direct=dealias_direct,
                              coordinates=coordinates)
        self.alpha = alpha  # Jacobi parameter
        self.beta = beta    # Jacobi parameter
        self.gn = 1         # Jacobi scaling function

    @property
    def is_jacobi(self):
        return True

    def L2_norm_sq(self, i):
        return h(self.alpha, self.beta, i, 0, gn=self.gn)

    def l2_norm_sq(self, i=None):
        if i is None:
            return sp.lambdify(n, h(self.alpha, self.beta, n, 0, gn=self.gn))(np.arange(self.N))
        return h(self.alpha, self.beta, i, 0, gn=self.gn)

    @staticmethod
    def bnd_values(k=0, alpha=None, beta=None, gn=None):
        from shenfun.jacobi.recursions import bnd_values
        return bnd_values(alpha, beta, k=k, gn=gn)

    def reference_domain(self):
        return (-1, 1)

    def unweighted_quadrature_weights(self):
        r"""Return quadrature weights for unweighted integrals

        .. math::

            \int_{-1}^{1} f dx \approx \sum_{j=0}^{N} f(x_j) w_j,

        where `w_j` are the quadrature weights.

        """
        xj, wj = self.points_and_weights()
        h = self.l2_norm_sq()
        uj = np.zeros(len(xj))
        b = self.vandermonde(xj)
        for j in range(len(xj)):
            for i in range(self.N):
                if i == 0:
                    fi = 2
                else:
                    fl = self.orthogonal_basis_function(i=i).as_poly().integrate()
                    fi = fl(1)-fl(-1)
                uj[j] += (b[j, i] / h[i]) * fi
            uj[j] *= wj[j]
        return uj

    def lagrange_poly(self, i):
        """Return i'th Lagrange polynomial of self

        Parameters
        ----------
        i : int
            For the i'th Lagrange polynomial

        """
        xj, wj = self.points_and_weights()
        h = self.l2_norm_sq()
        b = self.vandermonde(xj)
        l = sp.S(0)
        for j in range(self.N):
            if j == 0:
                fl = 1
            else:
                fl = self.orthogonal_basis_function(i=j).as_poly()
            l += (b[i, j] / h[j]) * fl
        l *= wj[i]
        return l

    def get_bc_space(self):
        if self._bc_space:
            return self._bc_space
        self._bc_space = BCGeneric(self.N, bc=self.bcs, domain=self.domain, alpha=self.alpha, beta=self.alpha)
        return self._bc_space

class Orthogonal(JacobiBase):
    r"""Function space for regular (orthogonal) Jacobi functions

    The orthogonal basis is

    .. math::

        \phi_k = P^{(\alpha,\beta)}_k, \quad k = 0, 1, \ldots, N-1,

    where :math:`P^{(\alpha,\beta)}_k` is the `Jacobi polynomial <https://en.wikipedia.org/wiki/Jacobi_polynomials>`_.
    The basis is orthogonal with weight :math:`(1-x)^{\alpha}(1+x)^{\beta}`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """

    def __init__(self, N, quad="JG", alpha=0, beta=0, domain=(-1, 1),
                 dtype=float, padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        JacobiBase.__init__(self, N, quad=quad, alpha=alpha, beta=beta, domain=domain, dtype=dtype,
                            padding_factor=padding_factor, dealias_direct=dealias_direct,
                            coordinates=coordinates)
        self.plan(int(N*padding_factor), 0, dtype, {})

    @staticmethod
    def family():
        return 'jacobi'

    def stencil_matrix(self, N=None):
        N = self.N if N is None else N
        return SparseMatrix({0: 1}, (N, N))

    def get_orthogonal(self, **kwargs):
        d = dict(quad=self.quad,
                 domain=self.domain,
                 dtype=self.dtype,
                 alpha=self.alpha,
                 beta=self.beta,
                 padding_factor=self.padding_factor,
                 dealias_direct=self.dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return Orthogonal(self.N, **d)

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        assert self.quad == "JG"
        points, weights = roots_jacobi(N, self.alpha, self.beta)
        if weighted is False:
            weights = self.unweighted_quadrature_weights()

        if map_true_domain is True:
            points = self.map_true_domain(points)
        return points, weights

    def jacobi(self, x, alpha, beta, N):
        mode = config['bases']['jacobi']['mode']
        V = np.zeros((x.shape[0], N))
        if mode == 'numpy':
            for n in range(N):
                V[:, n] = eval_jacobi(n, alpha, beta, x)
        else:
            for n in range(N):
                V[:, n] = sp.lambdify(xp, sp.jacobi(n, alpha, beta, xp), 'mpmath')(x)
        return V

    def derivative_jacobi(self, x, alpha, beta, k=1):
        V = self.jacobi(x, alpha+k, beta+k, self.N)
        if k > 0:
            Vc = np.zeros_like(V)
            for j in range(k, self.N):
                dj = np.prod(np.array([j+alpha+beta+1+i for i in range(k)]))
                #dj = gamma(j+alpha+beta+1+k) / gamma(j+alpha+beta+1)
                Vc[:, j] = (dj/2**k)*V[:, j-k]
            V = Vc
        return V

    def vandermonde(self, x):
        return self.jacobi(x, self.alpha, self.beta, self.shape(False))

    @property
    def is_orthogonal(self):
        return True

    @staticmethod
    def short_name():
        return 'J'

    def orthogonal_basis_function(self, i=0, x=xp):
        return sp.jacobi(i, self.alpha, self.beta, x)

    def evaluate_basis(self, x, i=0, output_array=None):
        mode = config['bases']['jacobi']['mode']
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape)
        if mode == 'numpy':
            output_array = eval_jacobi(i, self.alpha, self.beta, x, out=output_array)
        else:
            f = self.orthogonal_basis_function(i, xp)
            output_array[:] = sp.lambdify(xp, f, 'mpmath')(x)
        return output_array

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        mode = config['bases']['jacobi']['mode']
        if x is None:
            x = self.points_and_weights(mode=mode)[0]
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)

        if mode == 'numpy':
            dj = np.prod(np.array([i+self.alpha+self.beta+1+j for j in range(k)]))
            output_array[:] = dj/2**k*eval_jacobi(i-k, self.alpha+k, self.beta+k, x)
        else:
            f = sp.jacobi(i, self.alpha, self.beta, xp)
            output_array[:] = sp.lambdify(xp, f.diff(xp, k), 'mpmath')(x)
        return output_array

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        mode = config['bases']['jacobi']['mode']
        if x is None:
            x = self.points_and_weights()[0]
        if mode == 'numpy':
            return self.derivative_jacobi(x, self.alpha, self.beta, k)
        else:
            N = self.shape(False)
            V = np.zeros((x.shape[0], N))
            for i in range(N):
                f = sp.jacobi(i, self.alpha, self.beta, xp)
                V[:, i] = sp.lambdify(xp, f.diff(xp, k), 'mpmath')(x)
        return V

    def evaluate_basis_all(self, x=None, argument=0):
        if x is None:
            x = self.points_and_weights()[0]
        return self.vandermonde(x)

    def eval(self, x, u, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.forward.output_array.dtype)
        x = self.map_reference_domain(x)
        P = self.vandermonde(x)
        output_array = np.dot(P, u, out=output_array)
        return output_array

    #def get_bc_space(self):
    #    if self._bc_space:
    #        return self._bc_space
    #    self._bc_space = BCGeneric(self.N, bc=self.bcs, domain=self.domain, alpha=self.alpha, beta=self.beta)
    #    return self._bc_space

CompositeBase = getCompositeBase(Orthogonal)
BCGeneric = getBCGeneric(CompositeBase)

class Phi1(CompositeBase):
    r"""Function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::
        \phi_k &= \frac{(1-x^2)}{h^{(1,\alpha,\beta)}_{k+1}} \frac{dP^{(\alpha,\beta)}_{k+1}}{dx}, \, k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= \frac{\alpha + 1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_0  - \frac{1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_1, \\
        \phi_{N-1} &= \frac{\beta + 1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_0+ \frac{1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_1,

    where

    .. math::

        h^{(1,\alpha,\beta)}_k&=\int_{-1}^1 \left(\frac{dP^{(\alpha,\beta)}_{k}}{dx}\right)^2 (1-x)^{\alpha+1}(1+x)^{\beta+1}dx, \\
            &= \frac{2^{\alpha+\beta+1}(\alpha+\beta+k+1) \Gamma{(\alpha+k+1)} \Gamma{(\beta+k+1)}}{(\alpha+\beta+2k+1) \Gamma{(k)} \Gamma{(\alpha+\beta+k+1)}},

    and the expansion is

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a \text{ and } u(1) = b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 2-tuple of numbers, optional
        Boundary conditions at, respectively, x=(-1, 1).
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        #self._stencil = {
        #   0: sp.simplify(b(alpha, beta, n+1, n) / h(alpha, beta, n, 0)),
        #   2: sp.simplify(b(alpha, beta, n+1, n+1) / h(alpha, beta, n+1, 0)),
        #   4: sp.simplify(b(alpha, beta, n+1, n+2) / h(alpha, beta, n+2, 0))}
        a, b = alpha, beta
        self._stencil = {
            0: 2**(-a-b)*sp.gamma(n+1)*sp.gamma(a+b+n+2)/((a+b+2*n+2)*sp.gamma(a+n+1)*sp.gamma(b+n+1)),
            2: -2**(-a-b)*sp.gamma(n+3)*sp.gamma(a+b+n+2)/((a+b+2*n+4)*sp.gamma(a+n+2)*sp.gamma(b+n+2))
        }
        if alpha != beta:
            self._stencil[1] = 2**(-a-b)*(a-b)*(a+b+2*n+3)*sp.gamma(n+2)*sp.gamma(a+b+n+2)/((a+b+2*n+2)*(a+b+2*n+4)*sp.gamma(a+n+2)*sp.gamma(b+n+2))

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    @staticmethod
    def short_name():
        return 'P1'


class Phi2(CompositeBase):
    r"""Function space for biharmonic equations

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-5` are

    .. math::
        \phi_k &= \frac{(1-x^2)^2}{h^{(2,\alpha,\beta)}_{k+2}} \frac{d^2P^{(\alpha,\beta)}_{k+2}}{dx^2},

    where

    .. math::

        h^{(2,\alpha,\beta)}_k&=\int_{-1}^1 \left(\frac{d^2P^{(\alpha,\beta)}_{k}}{dx^2}\right)^2 (1-x)^{\alpha+2}(1+x)^{\beta+2}dx, \\
            &= \frac{2^{\alpha+\beta+1}(\alpha+\beta+k+1) \Gamma{(\alpha+k+1)} \Gamma{(\beta+k+1)}}{(\alpha+\beta+2k+1) \Gamma{(k)} \Gamma{(\alpha+\beta+k+1)}},

    The 4 boundary basis functions are computed using :func:`.jacobi.findbasis.get_bc_basis`,
    but they are too messy to print here. We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(-1)=b, u(1)=c \text{ and } u'(1) = d.

    The last 4 basis functions are for boundary conditions and only used if
    a, b, c or d are different from 0. In one dimension :math:`\hat{u}_{N-4}=a`,
    :math:`\hat{u}_{N-3}=b`, :math:`\hat{u}_{N-2}=c` and :math:`\hat{u}_{N-1}=d`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 4-tuple of numbers, optional
        Boundary conditions.
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """

    def __init__(self, N, quad="JG", bc=(0, 0, 0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        a, b = alpha, beta
        #self._stencil = {
        #   0: sp.simplify(matpow(b, 2, alpha, beta, n+2, n) / h(alpha, beta, n, 0)),
        #   2: sp.simplify(matpow(b, 2, alpha, beta, n+2, n+1) / h(alpha, beta, n+1, 0)),
        #   4: sp.simplify(matpow(b, 2, alpha, beta, n+2, n+2) / h(alpha, beta, n+2, 0)),
        #   6: sp.simplify(matpow(b, 2, alpha, beta, n+2, n+3) / h(alpha, beta, n+3, 0)),
        #   8: sp.simplify(matpow(b, 2, alpha, beta, n+2, n+4) / h(alpha, beta, n+4, 0))}
        self._stencil = {
            0: 2**(-a - b + 1)*sp.gamma(n + 1)*sp.gamma(a + b + n + 3)/((a + b + 2*n + 2)*(a + b + 2*n + 3)*(a + b + 2*n + 4)*sp.gamma(a + n + 1)*sp.gamma(b + n + 1)),
            2: 2**(-a - b + 1)*(a + b + 2*n + 5)*(a**2 - 4*a*b - 2*a*n - 5*a + b**2 - 2*b*n - 5*b - 2*n**2 - 10*n - 12)*sp.gamma(n + 3)*sp.gamma(a + b + n + 3)/((a + b + 2*n + 3)*(a + b + 2*n + 4)*(a + b + 2*n + 6)*(a + b + 2*n + 7)*sp.gamma(a + n + 3)*sp.gamma(b + n + 3)),
            4: 2**(-a - b + 1)*sp.gamma(n + 5)*sp.gamma(a + b + n + 3)/((a + b + 2*n + 6)*(a + b + 2*n + 7)*(a + b + 2*n + 8)*sp.gamma(a + n + 3)*sp.gamma(b + n + 3))
        }
        if self.alpha != self.beta:
            self._stencil[1] = 2**(-a - b + 2)*(a - b)*sp.gamma(n + 2)*sp.gamma(a + b + n + 3)/((a + b + 2*n + 2)*(a + b + 2*n + 4)*(a + b + 2*n + 6)*sp.gamma(a + n + 2)*sp.gamma(b + n + 2))
            self._stencil[3] = -2**(-a - b + 2)*(a - b)*sp.gamma(n + 4)*sp.gamma(a + b + n + 3)/((a + b + 2*n + 4)*(a + b + 2*n + 6)*(a + b + 2*n + 8)*sp.gamma(a + n + 3)*sp.gamma(b + n + 3))

    @staticmethod
    def boundary_condition():
        return 'Biharmonic'

    @staticmethod
    def short_name():
        return 'P2'

class Phi3(CompositeBase):
    r"""Function space for 6th order equations

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-7` are

    .. math::
        \phi_k &= \frac{(1-x^2)^3}{h^{(3,\alpha,\beta)}_{k+3}} \frac{d^3P^{(\alpha,\beta)}_{k+3}}{dx^3}, \\

    where

    .. math::

        h^{(3,\alpha,\beta)}_k&=\int_{-1}^1 \left(\frac{d^3P^{(\alpha,\beta)}_{k}}{dx^3}\right)^2 (1-x)^{\alpha+3}(1+x)^{\beta+3}dx, \\

    The 6 boundary basis functions are computed using :func:`.jacobi.findbasis.get_bc_basis`,
    but they are too messy to print here. We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(-1)=b, u''(-1)=c, u(1)=ed u'(1)=e, u''(1)=f.

    The last 6 basis functions are for boundary conditions and only used if there
    are nonzero boundary conditions.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 6-tuple of numbers, optional
        Boundary conditions.
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(0,)*6, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        #self._stencil = {
        #   0: sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n) / h(self.alpha, self.beta, n, 0)),
        #   2: sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n+2) / h(self.alpha, self.beta, n+2, 0)),
        #   4: sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n+4) / h(self.alpha, self.beta, n+4, 0)),
        #   6: sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n+6) / h(self.alpha, self.beta, n+6, 0))}
        a, b = alpha, beta
        self._stencil = {
            0: 2**(-a - b + 2)*sp.gamma(n + 1)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 2)*(a + b + 2*n + 3)*(a + b + 2*n + 4)*(a + b + 2*n + 5)*(a + b + 2*n + 6)*sp.gamma(a + n + 1)*sp.gamma(b + n + 1)),
            2: 3*2**(-a - b + 2)*(a**2 - 3*a*b - a*n - 3*a + b**2 - b*n - 3*b - n**2 - 6*n - 8)*sp.gamma(n + 3)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 3)*(a + b + 2*n + 4)*(a + b + 2*n + 6)*(a + b + 2*n + 8)*(a + b + 2*n + 9)*sp.gamma(a + n + 3)*sp.gamma(b + n + 3)),
            4: 3*2**(-a - b + 2)*(-a**2 + 3*a*b + a*n + 4*a - b**2 + b*n + 4*b + n**2 + 8*n + 15)*sp.gamma(n + 5)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*sp.gamma(a + n + 4)*sp.gamma(b + n + 4)),
            6: -2**(-a - b + 2)*sp.gamma(n + 7)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 8)*(a + b + 2*n + 9)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*(a + b + 2*n + 12)*sp.gamma(a + n + 4)*sp.gamma(b + n + 4))
        }
        if alpha != beta:
            #self._stencil[1] = sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n+1) / h(self.alpha, self.beta, n+1, 0))
            #self._stencil[3] = sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n+3) / h(self.alpha, self.beta, n+3, 0))
            #self._stencil[5] = sp.simplify(matpow(b, 3, self.alpha, self.beta, n+3, n+5) / h(self.alpha, self.beta, n+5, 0))
            self._stencil[1] = 3*2**(-a - b + 2)*(a - b)*sp.gamma(n + 2)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 2)*(a + b + 2*n + 4)*(a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 8)*sp.gamma(a + n + 2)*sp.gamma(b + n + 2))
            self._stencil[3] = 2**(-a - b + 2)*(a - b)*(a + b + 2*n + 7)*(a**2 - 8*a*b - 6*a*n - 21*a + b**2 - 6*b*n - 21*b - 6*n**2 - 42*n - 70)*sp.gamma(n + 4)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 4)*(a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 8)*(a + b + 2*n + 9)*(a + b + 2*n + 10)*sp.gamma(a + n + 4)*sp.gamma(b + n + 4))
            self._stencil[5] = 3*2**(-a - b + 2)*(a - b)*sp.gamma(n + 6)*sp.gamma(a + b + n + 4)/((a + b + 2*n + 6)*(a + b + 2*n + 8)*(a + b + 2*n + 9)*(a + b + 2*n + 10)*(a + b + 2*n + 12)*sp.gamma(a + n + 4)*sp.gamma(b + n + 4))

    @staticmethod
    def boundary_condition():
        return '6th order'

    @staticmethod
    def short_name():
        return 'P3'


class Phi4(CompositeBase):
    r"""Function space for 8th order equations

    The basis functions :math:`\phi_k` for :math:`k=0, 1, \ldots, N-9` are

    .. math::
        \phi_k &= \frac{(1-x^2)^4}{h^{(4,\alpha,\beta)}_{k+4}} \frac{d^4P^{(\alpha,\beta)}_{k+4}}{dx^4}, \\

    where

    .. math::

        h^{(4,\alpha,\beta)}_k&=\int_{-1}^1 \left(\frac{d^4P^{(\alpha,\beta)}_{k}}{dx^4}\right)^2 (1-x)^{\alpha+4}(1+x)^{\beta+4}dx, \\
            &=\frac{2^{\alpha + \beta + 1} \left(\alpha + \beta + n + 1\right) \left(\alpha + \beta + n + 2\right) \left(\alpha + \beta + n + 3\right) \left(\alpha + \beta + n + 4\right) \Gamma\left(\alpha + n + 1\right) \Gamma\left(\beta + n + 1\right)}{\left(\alpha + \beta + 2 n + 1\right) \Gamma\left(n - 3\right) \Gamma\left(\alpha + \beta + n + 1\right)},

    The 8 boundary basis functions are computed using :func:`.jacobi.findbasis.get_bc_basis`,
    but they are too messy to print here. We have

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a, u'(-1)=b, u''(-1)=c, u''''(-1)=d, u(1)=e, u'(1)=f, u''(1)=g, u''''(1)=h.

    The last 8 basis functions are for boundary conditions and only used if there
    are nonzero boundary conditions.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 8-tuple of numbers, optional
        Boundary conditions.
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(0,)*8, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        #self._stencil = {
        #   0: sp.simplify(matpow(b, 4, alpha, beta, n+4, n) / h(alpha, beta, n, 0)),
        #   2: sp.simplify(matpow(b, 4, alpha, beta, n+4, n+2) / h(alpha, beta, n+2, 0)),
        #   4: sp.simplify(matpow(b, 4, alpha, beta, n+4, n+4) / h(alpha, beta, n+4, 0)),
        #   6: sp.simplify(matpow(b, 4, alpha, beta, n+4, n+6) / h(alpha, beta, n+6, 0)),
        #   8: sp.simplify(matpow(b, 4, alpha, beta, n+4, n+8) / h(alpha, beta, n+8, 0))}
        a, b = alpha, beta
        self._stencil = {
            0: 2**(-a - b + 3)*sp.gamma(n + 1)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 2)*(a + b + 2*n + 3)*(a + b + 2*n + 4)*(a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 7)*(a + b + 2*n + 8)*sp.gamma(a + n + 1)*sp.gamma(b + n + 1)),
            2: 2**(-a - b + 4)*(3*a**2 - 8*a*b - 2*a*n - 7*a + 3*b**2 - 2*b*n - 7*b - 2*n**2 - 14*n - 20)*sp.gamma(n + 3)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 3)*(a + b + 2*n + 4)*(a + b + 2*n + 6)*(a + b + 2*n + 7)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*sp.gamma(a + n + 3)*sp.gamma(b + n + 3)),
            4: 2**(-a - b + 3)*(a + b + 2*n + 9)*(a**4 - 16*a**3*b - 12*a**3*n - 54*a**3 + 36*a**2*b**2 + 24*a**2*b*n + 108*a**2*b - 6*a**2*n**2 - 54*a**2*n - 109*a**2 - 16*a*b**3 + 24*a*b**2*n + 108*a*b**2 + 48*a*b*n**2 + 432*a*b*n + 932*a*b + 12*a*n**3 + 162*a*n**2 + 714*a*n + 1026*a + b**4 - 12*b**3*n - 54*b**3 - 6*b**2*n**2 - 54*b**2*n - 109*b**2 + 12*b*n**3 + 162*b*n**2 + 714*b*n + 1026*b + 6*n**4 + 108*n**3 + 714*n**2 + 2052*n + 2160)*sp.gamma(n + 5)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 7)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*(a + b + 2*n + 12)*(a + b + 2*n + 13)*sp.gamma(a + n + 5)*sp.gamma(b + n + 5)),
            6: 2**(-a - b + 4)*(3*a**2 - 8*a*b - 2*a*n - 11*a + 3*b**2 - 2*b*n - 11*b - 2*n**2 - 22*n - 56)*sp.gamma(n + 7)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 7)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*(a + b + 2*n + 12)*(a + b + 2*n + 14)*(a + b + 2*n + 15)*sp.gamma(a + n + 5)*sp.gamma(b + n + 5)),
            8: 2**(-a - b + 3)*sp.gamma(n + 9)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 10)*(a + b + 2*n + 11)*(a + b + 2*n + 12)*(a + b + 2*n + 13)*(a + b + 2*n + 14)*(a + b + 2*n + 15)*(a + b + 2*n + 16)*sp.gamma(a + n + 5)*sp.gamma(b + n + 5))
        }

        if alpha != beta:
            #self._stencil[1] = sp.simplify(matpow(b, 4, alpha, beta, n+4, n+1) / h(alpha, beta, n+1, 0))
            #self._stencil[3] = sp.simplify(matpow(b, 4, alpha, beta, n+4, n+3) / h(alpha, beta, n+3, 0))
            #self._stencil[5] = sp.simplify(matpow(b, 4, alpha, beta, n+4, n+5) / h(alpha, beta, n+5, 0))
            #self._stencil[7] = sp.simplify(matpow(b, 4, alpha, beta, n+4, n+7) / h(alpha, beta, n+7, 0))
            self._stencil[1] = 2**(-a - b + 5)*(a - b)*sp.gamma(n + 2)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 2)*(a + b + 2*n + 4)*(a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 7)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*sp.gamma(a + n + 2)*sp.gamma(b + n + 2))
            self._stencil[3] = 2**(-a - b + 5)*(a - b)*(a**2 - 5*a*b - 3*a*n - 12*a + b**2 - 3*b*n - 12*b - 3*n**2 - 24*n - 43)*sp.gamma(n + 4)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 4)*(a + b + 2*n + 5)*(a + b + 2*n + 6)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*(a + b + 2*n + 12)*sp.gamma(a + n + 4)*sp.gamma(b + n + 4))
            self._stencil[5] = 2**(-a - b + 5)*(a - b)*(-a**2 + 5*a*b + 3*a*n + 15*a - b**2 + 3*b*n + 15*b + 3*n**2 + 30*n + 70)*sp.gamma(n + 6)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 6)*(a + b + 2*n + 7)*(a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 12)*(a + b + 2*n + 13)*(a + b + 2*n + 14)*sp.gamma(a + n + 5)*sp.gamma(b + n + 5))
            self._stencil[7] = -2**(-a - b + 5)*(a - b)*sp.gamma(n + 8)*sp.gamma(a + b + n + 5)/((a + b + 2*n + 8)*(a + b + 2*n + 10)*(a + b + 2*n + 11)*(a + b + 2*n + 12)*(a + b + 2*n + 13)*(a + b + 2*n + 14)*(a + b + 2*n + 16)*sp.gamma(a + n + 5)*sp.gamma(b + n + 5))

    @staticmethod
    def boundary_condition():
        return 'Biharmonic*2'

    @staticmethod
    def short_name():
        return 'P4'


class CompactDirichlet(CompositeBase):
    r"""Function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::
        \phi_k &= P^{(\alpha,\beta)}_k + a_k P^{(\alpha,\beta)}_{k+1} + b_k P^{(\alpha,\beta)}_{k+2}  \quad k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= \frac{\alpha + 1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_0  - \frac{1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_1, \\
        \phi_{N-1} &= \frac{\beta + 1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_0+ \frac{1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_1,

    where

    .. math::
        a_k &= \frac{\left(\alpha - \beta\right) \left(n + 1\right) \left(\alpha + \beta + 2 n + 3\right)}{\left(\alpha + n + 1\right) \left(\beta + n + 1\right) \left(\alpha + \beta + 2 n + 4\right)}, \\
        b_k &= - \frac{\left(n + 1\right) \left(n + 2\right) \left(\alpha + \beta + 2 n + 2\right)}{\left(\alpha + n + 1\right) \left(\beta + n + 1\right) \left(\alpha + \beta + 2 n + 4\right)}, \\

    and the expansion is

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) &= a \text{ and } u(1) = b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 2-tuple of numbers, optional
        Boundary conditions at, respectively, x=(-1, 1).
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        a, b = alpha, beta
        self._stencil = {
            0: 1,
            2: -(n + 1)*(n + 2)*(a + b + 2*n + 2)/((a + n + 1)*(b + n + 1)*(a + b + 2*n + 4))
        }
        if self.alpha != self.beta:
            self._stencil[1] = (a - b)*(n + 1)*(a + b + 2*n + 3)/((a + n + 1)*(b + n + 1)*(a + b + 2*n + 4))

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    @staticmethod
    def short_name():
        return 'CD'


class CompactNeumann(CompositeBase):
    r"""Function space for Neumann boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::
        \phi_k &= P^{(\alpha,\beta)}_k + a_k P^{(\alpha,\beta)}_{k+1} + b_k P^{(\alpha,\beta)}_{k+2}  \quad k=0, 1, \ldots, N-3, \\
        \phi_{N-2} &= \frac{\alpha + 1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_0  - \frac{1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_1, \\
        \phi_{N-1} &= \frac{\beta + 1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_0+ \frac{1}{\alpha + \beta + 2}P^{(\alpha,\beta)}_1,

    where

    .. math::
        a_k &= \frac{n \left(\alpha - \beta\right) \left(\alpha + \beta + n + 1\right) \left(\alpha + \beta + 2 n + 3\right)}{\left(\alpha + n + 1\right) \left(\beta + n + 1\right) \left(\alpha + \beta + n + 2\right) \left(\alpha + \beta + 2 n + 4\right)}, \\
        b_k &= - \frac{n \left(n + 1\right) \left(\alpha + \beta + n + 1\right) \left(\alpha + \beta + 2 n + 2\right)}{\left(\alpha + n + 1\right) \left(\beta + n + 1\right) \left(\alpha + \beta + n + 3\right) \left(\alpha + \beta + 2 n + 4\right)}, \\

    and the expansion is

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u'(-1) &= a \text{ and } u'(1) = b.

    The last two bases are for boundary conditions and only used if a or b are
    different from 0. In one dimension :math:`\hat{u}_{N-2}=a` and
    :math:`\hat{u}_{N-1}=b`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 2-tuple of numbers, optional
        Boundary conditions at, respectively, x=(-1, 1).
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(0, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            bc = BoundaryConditions({'left': {'N': bc[0]}, 'right': {'N': bc[1]}}, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        a, b = self.alpha, self.beta
        self._stencil = {
            0: 1,
            2: -n*(n+1)*(a+b+n+1)*(a+b+2*n+2)/((a+n+1)*(b+n+1)*(a+b+n+3)*(a+b+2*n+4))
        }
        if self.alpha != self.beta:
            self._stencil[1] = n*(a-b)*(a+b+n+1)*(a+b+2*n+3)/((a+n+1)*(b+n+1)*(a+b+n+2)*(a+b+2*n+4))

    @staticmethod
    def boundary_condition():
        return 'Neumann'

    @staticmethod
    def short_name():
        return 'CN'


class UpperDirichlet(CompositeBase):
    r"""Function space for Dirichlet boundary conditions at right edge

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::
        \phi_k &= P^{(\alpha, \beta)}_k - \frac{(k+1)}{\alpha+k+1} P^{(\alpha,\beta)}_{k+1}  \quad k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= 1

    and the expansion is

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(1) = a.

    The last basis function is for boundary condition and only used if a is
    different from 0. In one dimension :math:`\hat{u}_{N-1}=a`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 2-tuple of (None, number), optional
        Boundary condition at x=1.
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(None, 0), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        self._stencil = {0: 1, 1: -(n+1)/(self.alpha+n+1)}

    @staticmethod
    def boundary_condition():
        return 'UpperDirichlet'

    @staticmethod
    def short_name():
        return 'UD'


class LowerDirichlet(CompositeBase):
    r"""Function space for Dirichlet boundary condition at left edge

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::
        \phi_k &= P^{(\alpha, \beta)}_k + \frac{(k+1)}{\beta+k+1} P^{(\alpha,\beta)}_{k+1}  \quad k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= 1

    and the expansion is

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(-1) = a.

    The last basis function is for boundary condition and only used if a is
    different from 0. In one dimension :math:`\hat{u}_{N-1}=a`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - JG - Jacobi-Gauss
    bc : 2-tuple of (number, None), optional
        Boundary condition at x=-1.
    alpha : number, optional
        Parameter of the Jacobi polynomial
    beta : number, optional
        Parameter of the Jacobi polynomial
    domain : 2-tuple of numbers, optional
        The computational domain
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`
    """
    def __init__(self, N, quad="JG", bc=(0, None), domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, alpha=0, beta=0,
                 coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)
        self._stencil = {0: 1, 1: (n+1)/(self.beta+n+1)}

    @staticmethod
    def boundary_condition():
        return 'LowerDirichlet'

    @staticmethod
    def short_name():
        return 'LD'


class Generic(CompositeBase):
    r"""Function space for space with any boundary conditions

    Any combination of Dirichlet and Neumann is possible.

    Parameters
    ----------
    N : int, optional
        Number of quadrature points
    quad : str, optional
        Type of quadrature
        - JG - Jacobi-Gauss
    bc : dict, optional
        The dictionary must have keys 'left' and 'right', to describe boundary
        conditions on the left and right boundaries, and a list of 2-tuples to
        specify the condition. Specify Dirichlet on both ends with

            {'left': {'D': a}, 'right': {'D': b}}

        for some values `a` and `b`, that will be neglected in the current
        function. Specify mixed Neumann and Dirichlet as

            {'left': {'N': a}, 'right': {'N': b}}

        For both conditions on the right do

            {'right': {'N': a, 'D': b}}

        Any combination should be possible, and it should also be possible to
        use second derivatives `N2`. See :class:`~shenfun.spectralbase.BoundaryConditions`.
    domain : 2-tuple of numbers, optional
        The computational domain
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    Note
    ----
    A test function is always using homogeneous boundary conditions.

    """
    def __init__(self, N, quad="JG", bc={}, domain=(-1, 1), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None,
                 alpha=0, beta=0, **kw):
        from shenfun.utilities.findbasis import get_stencil_matrix
        self._stencil = get_stencil_matrix(bc, 'jacobi', alpha, beta, 1)
        if not isinstance(bc, BoundaryConditions):
            bc = BoundaryConditions(bc, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               alpha=alpha, beta=beta, coordinates=coordinates)

    @staticmethod
    def boundary_condition():
        return 'Generic'

    @staticmethod
    def short_name():
        return 'GJ'



from shenfun import *
import sympy as sp

x = sp.Symbol('x', real=True)
N = 100
omegax = sp.exp(sp.loggamma(x + sp.S.Half) - sp.loggamma(x + 1))
T = FunctionSpace(N, 'C')
L = FunctionSpace(N, 'L')
M = inner(TestFunction(T), TrialFunction(L))
k0 = 26
T0 = FunctionSpace(25, 'C', domain=[k0, N-1])
m0 = Array(T0, buffer=omegax)
T1 = FunctionSpace(25, 'C', domain=[k0-2, N-3])
m1 = Array(T1, buffer=omegax)
T2 = FunctionSpace(25, 'C', domain=[k0-4, N-5])
m2 = Array(T2, buffer=omegax)



from numbers import Number
import sympy as sp
import numpy as np
from shenfun.spectralbase import inner_product, SpectralBase, MixedFunctionSpace
from shenfun.matrixbase import TPMatrix
from shenfun.tensorproductspace import TensorProductSpace, CompositeSpace
from shenfun.utilities import dx, split, scalar_product, CachedArrayDict
from shenfun.config import config
from .arguments import Expr, Function, BasisFunction, Array, TestFunction

__all__ = ('inner', 'Inner')

#pylint: disable=line-too-long,inconsistent-return-statements,too-many-return-statements


def inner(expr0, expr1, output_array=None, assemble=None, kind=None, fixed_resolution=None, return_matrices=False):
    r"""
    Return (weighted or unweighted) discrete inner product

    .. math::

        (f, g)_w^N = \sum_{i\in\mathcal{I}}f(x_i) \overline{g}(x_i) w_i \approx \int_{\Omega} f\, \overline{g}\, w\, dx

    where :math:`\mathcal{I}=0, 1, \ldots, N-1, N \in \mathbb{Z}^+`, :math:`f`
    is a number or an expression linear in a :class:`.TestFunction`,
    and :math:`g` is an expression that is linear in :class:`.TrialFunction`
    or :class:`.Function`, or it is simply an :class:`.Array` (a solution interpolated on the
    quadrature mesh in physical space). :math:`w` is a weight associated with
    chosen basis, and :math:`w_i` are quadrature weights.

    Note
    ----
        If :math:`f` is a number (typically one) and :math:`g` an :class:`.Array`, then `inner`
        represents an unweighted, regular integral over the domain.

    If the expressions are created in a multidimensional :class:`.TensorProductSpace`,
    then the sum above is over all dimensions. In 2D it becomes:

    .. math::

        (f, g)_w^N = \sum_{i\in\mathcal{I}}\sum_{j\in\mathcal{J}} f(x_i, y_j) \overline{g}(x_i, y_j) w_j w_i

    where :math:`\mathcal{J}=0, 1, \ldots, M, M \in \mathbb{Z}^+`.

    Parameters
    ----------
    expr0, expr1 : :class:`.Expr`, :class:`.BasisFunction`, :class:`.Array`, number
        Either one can be an expression involving a
        BasisFunction (:class:`.TestFunction`, :class:`.TrialFunction` or
        :class:`.Function`) an Array or a number. With expressions (Expr) on a
        BasisFunction we typically mean terms like div(u) or grad(u), where
        u is any one of the different types of BasisFunction.
        If one of ``expr0``/``expr1`` involves a TestFunction and the other one
        involves a TrialFunction, then a tensor product matrix (or a list of
        tensor product matrices) is returned.
        If one of ``expr0``/``expr1`` involves a TestFunction and the other one
        involves a Function, or a plain Array, then a linear form is assembled
        and a Function is returned.
        If one of ``expr0``/``expr1`` is a number (typically 1), or a tuple of
        numbers for a vector, then the inner product represents a non-weighted
        integral over the domain. If a single number, then the other expression
        must be a scalar Array or a Function. If a tuple of numbers, then the
        Array/Function must be a vector.

    output_array : Function
        Optional return array for linear form.

    assemble : None or str, optional
        Determines how to perform the integration

        - 'quadrature' (default)
        - 'exact'
        - 'adaptive'

        Exact and adaptive should result in the same matrix. Exact computes the
        integral using `Sympy integrate <https://docs.sympy.org/latest/modules/integrals/integrals.html>`_,
        whereas adaptive makes use of adaptive quadrature through `scipy <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.quadrature.html>`_.
        For most bilinear forms quadrature will also return the same matrix as exact.
        With non-constant measures (like for curvilinear coordinates) there is more
        likely to be differences.

    kind : None, str or dict, optional
        Alternative methods.

        For bilinear forms (matrices) kind is a string:

        - 'implemented' - Hardcoded implementations
        - 'stencil' - Use orthogonal bases and stencil-matrices
        - 'vandermonde' - Use Vandermonde matrix

        The default (for kind=None) is to first try to look for implemented kind,
        and if that fails try first 'stencil' and then finally fall back on
        vandermonde. Vandermonde creates a dense matrix of size NxN, so it should
        be avoided (e.g., by implementing the matrix) for large N.

        For linear forms (vectors) the kind keyword can be used to overload
        the default methods for transforms set in config['transforms']['kind'].
        Hence, kind is a dictionary with family as key and values either one of
        the possible methods

        - 'fast' - Use FFT (only Fourier and Chebyshev first and second kind)
        - 'recursive' - Low-memory implementation (only for Jacobi polynomials)
        - 'vandermonde' - Use Vandermonde matrix

        E.g., kind={'chebyshev': 'recursive'}.
        Note that for one-dimensional problems it is enough to use just the
        value of the dictionary.

    fixed_resolution : Number or sequence of integers, optional
        A fixed number of quadrature points used to compute the inner product
        along each dimension of the domain. If 'fixed_resolution' is set, then
        assemble is set to 'quadrature' and kind is set to 'vandermonde'.
        fixed_resolution is argument to :meth:`.TensorProductSpace.get_refined`.

    return_matrices : bool, optional
        For linear forms, whether to simply return the matrices that are used
        to compute the result with matrix-vector products.

    Note
    ----
    For most matrices all methods will lead to the same result. For bilinear
    forms with polynomial coefficients regular quadrature will become
    inaccurate and a 'fixed_resolution' higher than the regular number of
    quadrature points should be considered.

    Returns
    -------
    Depending on dimensionality and the arguments to the forms

        :class:`.Function`
        for linear forms.

        :class:`.SparseMatrix`
        for bilinear 1D forms.

        :class:`.TPMatrix` or list of :class:`.TPMatrix`
        for bilinear multidimensional forms.

        Number, for non-weighted integral where either one of the arguments
        is a number.

    See Also
    --------
    :func:`.project`

    Example
    -------
    Compute mass matrix of Shen's Chebyshev Dirichlet basis:

    >>> from shenfun import FunctionSpace, TensorProductSpace
    >>> from shenfun import TestFunction, TrialFunction, Array
    >>> SD = FunctionSpace(6, 'Chebyshev', bc=(0, 0))
    >>> u = TrialFunction(SD)
    >>> v = TestFunction(SD)
    >>> B = inner(v, u)
    >>> d = {-2: np.array([-np.pi/2]),
    ...       0: np.array([ 1.5*np.pi, np.pi, np.pi, np.pi]),
    ...       2: np.array([-np.pi/2])}
    >>> [np.all(abs(B[k]-v) < 1e-7) for k, v in d.items()]
    [True, True, True]

    # Compute unweighted integral
    >>> F = FunctionSpace(10, 'F', domain=(0, 2*np.pi))
    >>> T = TensorProductSpace(comm, (SD, F))
    >>> area = inner(1, Array(T, val=1))
    >>> print('Area of domain =', area)
    Area of domain = 12.56637061435917

    """
    # Wrap a pure numpy array in Array
    if isinstance(expr0, np.ndarray) and not isinstance(expr0, (Array, Function)):
        assert isinstance(expr1, (Expr, BasisFunction))
        if not expr0.flags['C_CONTIGUOUS']:
            expr0 = expr0.copy()
        expr0 = Array(expr1.function_space(), buffer=expr0)
    if isinstance(expr1, np.ndarray) and not isinstance(expr1, (Array, Function)):
        assert isinstance(expr0, (Expr, BasisFunction))
        if not expr1.flags['C_CONTIGUOUS']:
            expr1 = expr1.copy()
        expr1 = Array(expr0.function_space(), buffer=expr1)

    if isinstance(expr0, Number):
        if isinstance(expr1, TestFunction):
            expr0 = sp.sympify(expr0)
        else:
            assert isinstance(expr1, (Array, Function))
            space = expr1.function_space()
            if isinstance(space, (TensorProductSpace, CompositeSpace)):
                df = np.prod(np.array([float(base.domain_factor()) for base in space.bases]))
            elif isinstance(space, SpectralBase):
                df = float(space.domain_factor())
            if isinstance(expr1, Function):
                #return (expr0/df)*dx(expr1.backward())
                expr1 = expr1.backward()
            if hasattr(space, 'hi'):
                if space.hi.prod() != 1:
                    expr1 = space.get_measured_array(expr1.copy())
            return (expr0/df)*dx(expr1)

    if isinstance(expr1, Number):
        if isinstance(expr0, TestFunction):
            expr1 = sp.sympify(expr1)
        else:
            assert isinstance(expr0, (Array, Function))
            space = expr0.function_space()
            if isinstance(space, (TensorProductSpace, CompositeSpace)):
                df = np.prod(np.array([float(base.domain_factor()) for base in space.bases]))
            elif isinstance(space, SpectralBase):
                df = float(space.domain_factor())
            if isinstance(expr0, Function):
                #return (expr1/df)*dx(expr0.backward())
                expr0 = expr0.backward()
            if hasattr(space, 'hi'):
                if space.hi.prod() != 1:
                    expr0 = space.get_measured_array(expr0.copy())
            return (expr1/df)*dx(expr0)

    if isinstance(expr0, tuple):
        assert isinstance(expr1, (Array, Function))
        space = expr1.function_space()
        assert isinstance(space, CompositeSpace)
        assert len(expr0) == len(space)
        result = 0.0
        for e0i, e1i in zip(expr0, expr1):
            result += inner(e0i, e1i)
        return result

    if isinstance(expr1, tuple):
        assert isinstance(expr0, (Array, Function))
        space = expr0.function_space()
        assert isinstance(space, CompositeSpace)
        assert len(expr1) == len(space)
        result = 0.0
        for e0i, e1i in zip(expr0, expr1):
            result += inner(e0i, e1i)
        return result

    if isinstance(expr0, sp.Expr) or isinstance(expr1, sp.Expr):
        # Special linear form using sympy expression
        if isinstance(expr0, sp.Expr):
            assert isinstance(expr1, TestFunction)
            test = expr1
            trial = expr0
        else:
            assert isinstance(expr0, TestFunction)
            test = expr0
            trial = expr1

        if output_array is None:
            output_array = Function(test.function_space())

        if assemble in ('exact', 'adaptive'):
            output_array = scalar_product(test, trial, output_array, assemble=assemble)

        else: # quadrature
            if fixed_resolution is not None:
                M = fixed_resolution
                testM = test.function_space().get_refined(M)
                outM = testM.scalar_product(Array(testM, buffer=trial), kind=kind)
                output_array[:test.function_space().dim()] = outM[:test.function_space().dim()]
            else:
                output_array = test.function_space().scalar_product(Array(test.function_space(), buffer=trial), output_array, kind=kind)
        return output_array

    assert np.all([hasattr(e, 'argument') for e in (expr0, expr1)])
    t0 = expr0.argument
    t1 = expr1.argument
    if t0 == 0:
        assert t1 in (1, 2)
        test = expr0
        trial = expr1
    elif t0 in (1, 2):
        assert t1 == 0
        test = expr1
        trial = expr0
    else:
        raise RuntimeError

    recursive = test.function_space().is_composite_space
    if isinstance(trial, Array):
        assert Expr(trial).expr_rank() == test.expr_rank()
    elif isinstance(trial, BasisFunction):
        recursive *= (trial.expr_rank() > 0)
    if test.expr_rank() == 0:
        recursive = False

    if recursive: # Use recursive algorithm for vector expressions of expr_rank > 0, e.g., inner(v, grad(u))
        gij = test.function_space().coors.get_metric_tensor(config['basisvectors'])
        if trial.argument == 2 and not return_matrices:
            # linear form
            if output_array is None:
                output_array = Function(test.function_space())

            w0 = np.zeros(test.function_space().flatten()[0].shape(), dtype=output_array.dtype)
            if test.tensor_rank == 2:

                for i, (tei, xi) in enumerate(zip(test, output_array)):
                    for j, (teij, xij) in enumerate(zip(tei, xi)):
                        for k, trk in enumerate(trial):
                            if gij[i, k] == 0:
                                continue
                            for l, trkl in enumerate(trk):
                                if gij[j, l] == 0:
                                    continue

                                w0.fill(0)
                                xij += inner(teij*gij[i, k]*gij[j, l], trkl, output_array=w0, assemble=assemble, kind=kind, fixed_resolution=fixed_resolution, return_matrices=return_matrices)

            elif test.tensor_rank == 1:
                for i, (te, x) in enumerate(zip(test, output_array)):
                    for j, tr in enumerate(trial):
                        if gij[i, j] == 0:
                            continue
                        w0.fill(0)
                        x += inner(te*gij[i, j], tr, output_array=w0, assemble=assemble, kind=kind, fixed_resolution=fixed_resolution, return_matrices=return_matrices)

            return output_array

        result = []

        if test.tensor_rank == 2:
            for i, tei in enumerate(test):
                for j, teij in enumerate(tei):
                    for k, trk in enumerate(trial):
                        if gij[i, k] == 0:
                            continue
                        for l, trkl in enumerate(trk):
                            if gij[j, l] == 0:
                                continue
                            p = inner(teij*gij[i, k]*gij[j, l], trkl, assemble=assemble, kind=kind, fixed_resolution=fixed_resolution, return_matrices=return_matrices)
                            result += p if isinstance(p, list) else [p]

        elif test.tensor_rank == 1:
            for i, te in enumerate(test):
                for j, tr in enumerate(trial):
                    if gij[i, j] == 0:
                        continue
                    l = inner(te, tr*gij[i, j], assemble=assemble, kind=kind, fixed_resolution=fixed_resolution, return_matrices=return_matrices)
                    result += l if isinstance(l, list) else [l]

        return result[0] if len(result) == 1 else result

    if output_array is None and trial.argument == 2:
        output_array = Function(test.function_space())

    if trial.argument > 1:
        # Linear form
        assert isinstance(test, (Expr, BasisFunction))
        assert test.argument == 0
        space = test.function_space()
        if isinstance(trial, Array):
            if trial.tensor_rank == 0 and isinstance(test, BasisFunction):
                output_array = space.scalar_product(trial, output_array, kind=kind)
                return output_array
            # project to orthogonal. Cannot use trial space, because the Array trial
            # may not fit with the boundary conditions of the trialspace
            orthogonal = trial.function_space().get_orthogonal()
            trial = Array(orthogonal, buffer=trial)
            trial = trial.forward()

    assert isinstance(trial, (Expr, BasisFunction))
    assert isinstance(test, (Expr, BasisFunction))

    if isinstance(trial, BasisFunction):
        trial = Expr(trial)
    if isinstance(test, BasisFunction):
        test = Expr(test)

    assert test.expr_rank() == trial.expr_rank()

    testspace = test.base.function_space()
    trialspace = trial.base.function_space()
    test_scale = test.scales()
    trial_scale = trial.scales()
    uh = None
    if trial.argument == 2:
        uh = trial.base

    gij = testspace.coors.get_metric_tensor(config['basisvectors'])

    A = []
    for vec_i, (base_test, test_ind) in enumerate(zip(test.terms(), test.indices())): # vector/scalar
        for vec_j, (base_trial, trial_ind) in enumerate(zip(trial.terms(), trial.indices())):
            g = 1 if len(test.terms()) == 1 else gij[vec_i, vec_j]
            if g == 0:
                continue
            for test_j, b0 in enumerate(base_test):              # second index test
                for trial_j, b1 in enumerate(base_trial):        # second index trial
                    dV = sp.simplify(test_scale[vec_i][test_j]*trial_scale[vec_j][trial_j]*testspace.coors.sg*g, measure=testspace.coors._measure)
                    dV = testspace.coors.refine(dV)

                    assert len(b0) == len(b1)
                    trial_sp = trialspace
                    if isinstance(trialspace, (CompositeSpace, MixedFunctionSpace)): # could operate on a vector, e.g., div(u), where u is vector
                        trial_sp = trialspace.flatten()[trial_ind[trial_j]]
                    test_sp = testspace
                    if isinstance(testspace, (CompositeSpace, MixedFunctionSpace)):
                        test_sp = testspace.flatten()[test_ind[test_j]]
                    has_bcs = False
                    # Check if scale is zero
                    if dV == 0:
                        continue
                    for dv in split(dV):
                        sc = dv['coeff']
                        M = []
                        DM = []
                        for i, (a, b) in enumerate(zip(b0, b1)): # Third index, one inner for each dimension
                            ts = trial_sp[i]
                            tt = test_sp[i]
                            msx = 'xyzrs'[i]
                            msi = dv[msx]

                            # assemble inner product
                            AA = inner_product((tt, a), (ts, b), msi, assemble=assemble, kind=kind, fixed_resolution=fixed_resolution)
                            if len(AA) == 0:
                                sc = 0
                                continue

                            #if not abs(AA.scale-1.) < 1e-8:
                            #    AA.incorporate_scale()
                            M.append(AA)

                            if ts.has_nonhomogeneous_bcs:
                                tsc = ts.get_bc_space()
                                BB = inner_product((tt, a), (tsc, b), msi, assemble=assemble, kind=kind, fixed_resolution=fixed_resolution)
                                #if not abs(BB.scale-1.) < 1e-8:
                                #    BB.incorporate_scale()
                                #if BB.diags('csr').nnz > 0:
                                if len(BB) > 0:
                                    DM.append(BB)
                                    has_bcs = True
                                else:
                                    DM.append(0)
                            else:
                                DM.append(AA)

                        if sc == 0:
                            continue
                        sc = tt.broadcast_to_ndims(np.array([sc]))

                        if len(M) == 1: # 1D case
                            M[0].global_index = (test_ind[test_j], trial_ind[trial_j])
                            M[0].scale = sc[0]
                            M[0].testbase = testspace
                            M[0].trialbase = trialspace
                            M[0].incorporate_scale()
                            A.append(M[0])
                        else:
                            A.append(TPMatrix(M, test_sp, trial_sp, sc, (test_ind[test_j], trial_ind[trial_j]), testspace, trialspace))
                        if has_bcs:
                            if len(DM) == 1: # 1D case
                                DM[0].global_index = (test_ind[test_j], trial_ind[trial_j])
                                DM[0].scale = sc
                                DM[0].testbase = testspace
                                DM[0].trialbase = testspace
                                A.append(DM[0])
                            else:
                                if len(trial_sp.get_nonhomogeneous_axes()) == 1:
                                    A.append(TPMatrix(DM, test_sp, trial_sp, sc, (test_ind[test_j], trial_ind[trial_j]), testspace, trialspace))
                                elif len(trial_sp.get_nonhomogeneous_axes()) == 2:
                                    if DM[1] != 0:
                                        A.append(TPMatrix([M[0], DM[1]], test_sp, trial_sp, sc, (test_ind[test_j], trial_ind[trial_j]), testspace, trialspace))
                                    if DM[0] != 0:
                                        A.append(TPMatrix([DM[0], M[1]], test_sp, trial_sp, sc, (test_ind[test_j], trial_ind[trial_j]), testspace, trialspace))
                                    if DM[0] != 0 and DM[1] != 0:
                                        A.append(TPMatrix(DM, test_sp, trial_sp, sc, (test_ind[test_j], trial_ind[trial_j]), testspace, trialspace))

    # At this point A contains all matrices of the form. The length of A is
    # the number of inner products. For each index into A there are ndim 1D
    # inner products along, e.g., x, y and z-directions, or just x, y for 2D.
    # The outer product of these matrices is a tensorproduct matrix, and we
    # store the matrices using the TPMatrix class.
    #
    # There are now two possibilities, either a linear or a bilinear form.
    # A linear form has trial.argument == 2, whereas a bilinear form has
    # trial.argument == 1. A linear form should assemble to an array and
    # return this array. A bilinear form, on the other hand, should return
    # matrices. Which matrices, and how many, will of course depend on the
    # form and the number of terms.

    # Bilinear form, return matrices
    if trial.argument == 1:
        return A[0] if len(A) == 1 else A

    if return_matrices:
        return A

    # Linear form, return output_array
    wh = np.zeros_like(output_array)
    for b in A:
        if uh.function_space().is_composite_space and wh.ndim == b.dimensions:
            wh = b.matvec(uh.v[b.global_index[1]], wh)
        elif uh.function_space().is_composite_space and wh.ndim > b.dimensions:
            wh[b.global_index[0]] = b.matvec(uh.v[b.global_index[1]], wh[b.global_index[0]])
        else:
            wh = b.matvec(uh, wh)
        output_array += wh
        wh.fill(0)
    return output_array

work = CachedArrayDict()

class Inner:
    """Return an instance of a class that can perform the inner product
    of the linear form efficiently through a matrix-vector product

    Parameters
    ----------
    v : :class:`.TestFunction`
    uh : Instance of either one of
        - :class:`.Expr`
        - :class:`.BasisFunction`

    Note
    ----
    This is an optimization only for linear forms, not bilinear.
    There is no need to use this class for regular scalar products, where `uh`
    is simply an Array.
    """
    def __init__(self, v, uh):
        from shenfun.matrixbase import get_simplified_tpmatrices
        assert isinstance(uh, (Expr, BasisFunction, Array))
        assert isinstance(v, TestFunction)
        assert uh.argument == 2
        self.uh = [uh]
        A = inner(v, uh, return_matrices=True)
        if isinstance(A[0], TPMatrix):
            A = get_simplified_tpmatrices(A)
        elif isinstance(A, Function):
            A = [A]
        self.A = [A]
        self.output_array = Function(v.function_space())

    def __call__(self):
        wh = work[(self.output_array, 0, True)]
        self.output_array.fill(0)
        for uh, A in zip(self.uh, self.A):
            uh = uh.base if uh.base is not None else uh
            for b in A:
                if isinstance(b, Function) and isinstance(uh, Array):
                    V = b.function_space()
                    wh = V.scalar_product(uh, wh)
                    self.output_array += wh
                elif uh.function_space().is_composite_space and wh.ndim == b.dimensions:
                    wh = b.matvec(uh.v[b.global_index[1]], wh)
                    self.output_array += wh
                elif uh.function_space().is_composite_space and wh.ndim > b.dimensions:
                    wh[b.global_index[0]] = b.matvec(uh.v[b.global_index[1]], wh[b.global_index[0]])
                    self.output_array.v[b.global_index[0]] += wh[b.global_index[0]]
                else:
                    wh = b.matvec(uh, wh)
                    self.output_array += wh
                wh.fill(0)
        return self.output_array

    def __add__(self, c):
        assert isinstance(c, Inner)
        assert c.output_array.function_space() == self.output_array.function_space()
        self.A += c.A
        self.uh += c.uh
        return self



# pylint: disable=line-too-long
import copy
import re
from numpy import dtype, array, invert, take

__all__ = ('generate_xdmf',)

xdmffile = """<?xml version="1.0" encoding="utf-8"?>
<Xdmf xmlns:xi="http://www.w3.org/2001/XInclude" Version="2.1">
  <Domain>
    <Grid Name="Structured Grid" GridType="Collection" CollectionType="Temporal">
      <Time TimeType="List"><DataItem Format="XML" Dimensions="{1}"> {0} </DataItem></Time>
      {2}
    </Grid>
  </Domain>
</Xdmf>
"""

def get_grid(geometry, topology, attrs):
    return """<Grid GridType="Uniform">
        {0}
        {1}
        {2}
      </Grid>
      """.format(geometry, topology, attrs)

def get_geometry(kind=0, dim=2):
    assert kind in (0, 1)
    assert dim in (2, 3)
    if dim == 2:
        if kind == 0:
            return """<Geometry Type="ORIGIN_DXDY">
          <DataItem Format="XML" NumberType="Float" Dimensions="2">
            {0} {1}
          </DataItem>
          <DataItem Format="XML" NumberType="Float" Dimensions="2">
            {2} {3}
          </DataItem>
        </Geometry>"""

        return """<Geometry Type="VXVYVZ">
          <DataItem Format="HDF" NumberType="Float" Precision="{0}" Dimensions="{1}">
            {3}:{6}/mesh/{4}
          </DataItem>
          <DataItem Format="HDF" NumberType="Float" Precision="{0}" Dimensions="{2}">
            {3}:{6}/mesh/{5}
          </DataItem>
          <DataItem Format="XML" NumberType="Float" Precision="8" Dimensions="1">
            0
          </DataItem>
        </Geometry>"""

    if dim == 3:
        if kind == 0:
            return """<Geometry Type="ORIGIN_DXDYDZ">
          <DataItem Format="XML" NumberType="Float" Dimensions="3">
            {0} {1} {2}
          </DataItem>
          <DataItem Format="XML" NumberType="Float" Dimensions="3">
            {3} {4} {5}
          </DataItem>
        </Geometry>"""

        return """<Geometry Type="VXVYVZ">
          <DataItem Format="HDF" NumberType="Float" Precision="{0}" Dimensions="{3}">
            {4}:{8}/mesh/{5}
          </DataItem>
          <DataItem Format="HDF" NumberType="Float" Precision="{0}" Dimensions="{2}">
            {4}:{8}/mesh/{6}
          </DataItem>
          <DataItem Format="HDF" NumberType="Float" Precision="{0}" Dimensions="{1}">
            {4}:{8}/mesh/{7}
          </DataItem>
        </Geometry>"""

def get_topology(dims, kind=0):
    assert len(dims) in (2, 3)
    co = 'Co' if kind == 0 else ''
    if len(dims) == 2:
        return """<Topology Dimensions="1 {0} {1}" Type="3D{2}RectMesh"/>""".format(dims[0], dims[1], co)
    if len(dims) == 3:
        return """<Topology Dimensions="{0} {1} {2}" Type="3D{3}RectMesh"/>""".format(dims[0], dims[1], dims[2], co)

def get_attribute(attr, h5filename, dims, prec):
    name = attr.split("/")[0]
    assert len(dims) in (2, 3)
    if len(dims) == 2:
        return """<Attribute Name="{0}" Center="Node">
          <DataItem Format="HDF" NumberType="Float" Precision="{5}" Dimensions="1 {1} {2}">
            {3}:/{4}
          </DataItem>
        </Attribute>
        """.format(name, dims[0], dims[1], h5filename, attr, prec)

    return """<Attribute Name="{0}" Center="Node">
          <DataItem Format="HDF" NumberType="Float" Precision="{6}" Dimensions="{1} {2} {3}">
            {4}:/{5}
          </DataItem>
        </Attribute>
        """.format(name, dims[0], dims[1], dims[2], h5filename, attr, prec)

def generate_xdmf(h5filename, periodic=True, order='visit'):
    """Generate XDMF-files

    Parameters
    ----------
    h5filename : str
        Name of hdf5-file that we want to decorate with xdmf
    periodic : bool or dim-sequence of bools, optional
        If true along axis i, assume data is periodic.
        Only affects the calculation of the domain size and only if the
        domain is given as 2-tuple of origin+dx.
    order : str
        ``paraview`` or ``visit``
        For some reason Paraview and Visit requires the mesh stored in
        opposite order in the XDMF-file for 2D slices. Make choice of
        order here.
    """
    import h5py
    f = h5py.File(h5filename, 'a')
    keys = []
    f.visit(keys.append)
    assert order.lower() in ('paraview', 'visit')

    # Find unique scalar groups of 2D and 3D datasets
    datasets = {2:{}, 3:{}}
    for key in keys:
        if f[key.split('/')[0]].attrs['rank'] > 0:
            continue
        if isinstance(f[key], h5py.Dataset):
            if not ('mesh' in key or 'domain' in key or 'Vector' in key):
                tstep = int(key.split("/")[-1])
                ndim = int(key.split("/")[1][0])
                if ndim in (2, 3):
                    ds = datasets[ndim]
                    if tstep in ds:
                        ds[tstep] += [key]
                    else:
                        ds[tstep] = [key]
    if periodic is True:
        periodic = [0]*5
    elif periodic is False:
        periodic = [1]*5
    else:
        assert isinstance(periodic, (tuple, list))
        periodic = list(array(invert(periodic), int))

    coor = ['x0', 'x1', 'x2', 'x3', 'x4']
    for ndim, dsets in datasets.items():
        timesteps = list(dsets.keys())
        per = copy.copy(periodic)
        if not timesteps:
            continue
        timesteps.sort(key=int)
        tt = ""
        for i in timesteps:
            tt += "%s " %i
        datatype = f[dsets[timesteps[0]][0]].dtype
        assert datatype.char not in 'FDG', "Cannot use generate_xdmf to visualize complex data."
        prec = 4 if datatype is dtype('float32') else 8
        xff = {}
        geometry = {}
        topology = {}
        attrs = {}
        grid = {}
        NN = {}
        for name in dsets[timesteps[0]]:
            group = name.split('/')[0]
            if 'slice' in name:
                slices = name.split("/")[2]
            else:
                slices = 'whole'
            cc = copy.copy(coor)
            if slices not in xff:
                xff[slices] = copy.copy(xdmffile)
                N = list(f[name].shape)
                kk = 0
                sl = 0
                if 'slice' in slices:
                    ss = slices.split("_")
                    ii = []
                    for i, sx in enumerate(ss):
                        if 'slice' in sx:
                            ii.append(i)
                        else:
                            if len(f[group].attrs.get('shape')) == 3:      # 2D slice in 3D domain
                                kk = i
                                sl = int(sx)
                                N.insert(i, 1)
                    cc = take(coor, ii)
                else:
                    ii = list(range(ndim))
                NN[slices] = N
                if 'domain' in f[group].keys():
                    if ndim == 2 and ('slice' not in slices or len(f[group].attrs.get('shape')) > 3):
                        geo = get_geometry(kind=0, dim=2)
                        assert len(ii) == 2
                        i, j = ii
                        if order.lower() == 'paraview':
                            data = [f[group+'/domain/{}'.format(coor[i])][0],
                                    f[group+'/domain/{}'.format(coor[j])][0],
                                    f[group+'/domain/{}'.format(coor[i])][1]/(N[0]-per[i]),
                                    f[group+'/domain/{}'.format(coor[j])][1]/(N[1]-per[j])]
                            geometry[slices] = geo.format(*data)
                        else:
                            data = [f[group+'/domain/{}'.format(coor[j])][0],
                                    f[group+'/domain/{}'.format(coor[i])][0],
                                    f[group+'/domain/{}'.format(coor[j])][1]/(N[0]-per[j]),
                                    f[group+'/domain/{}'.format(coor[i])][1]/(N[1]-per[i])]
                            geometry[slices] = geo.format(*data)
                    else:
                        if ndim == 2:
                            ii.insert(kk, kk)
                            per[kk] = 0
                        i, j, k = ii
                        geo = get_geometry(kind=0, dim=3)
                        data = [f[group+'/domain/{}'.format(coor[i])][0],
                                f[group+'/domain/{}'.format(coor[j])][0],
                                f[group+'/domain/{}'.format(coor[k])][0],
                                f[group+'/domain/{}'.format(coor[i])][1]/(N[0]-per[i]),
                                f[group+'/domain/{}'.format(coor[j])][1]/(N[1]-per[j]),
                                f[group+'/domain/{}'.format(coor[k])][1]/(N[2]-per[k])]
                        if ndim == 2:
                            origin, dx = f[group+'/domain/x{}'.format(kk)]
                            M = f[group].attrs.get('shape')
                            pos = origin+dx/(M[kk]-per[kk])*sl
                            data[kk] = pos
                            data[kk+3] = pos
                        geometry[slices] = geo.format(*data)
                    topology[slices] = get_topology(N, kind=0)
                elif 'mesh' in f[group].keys():
                    if ndim == 2 and ('slice' not in slices or len(f[group].attrs.get('shape')) > 3):
                        geo = get_geometry(kind=1, dim=2)
                    else:
                        geo = get_geometry(kind=1, dim=3)

                    if ndim == 2 and ('slice' not in slices or len(f[group].attrs.get('shape')) > 3):
                        if order.lower() == 'paraview':
                            sig = (prec, N[0], N[1], h5filename, cc[0], cc[1], group)
                        else:
                            sig = (prec, N[1], N[0], h5filename, cc[1], cc[0], group)
                    else:
                        if ndim == 2: # 2D slice in 3D domain
                            pos = f[group+"/mesh/x{}".format(kk)][sl]
                            z = re.findall(r'<DataItem(.*?)</DataItem>', geo, re.DOTALL)
                            geo = geo.replace(z[2-kk], ' Format="XML" NumberType="Float" Precision="{0}" Dimensions="{%d}">\n           {%d}\n          '%(1+kk, 7-kk))
                            cc = list(cc)
                            cc.insert(kk, pos)
                        sig = (prec, N[0], N[1], N[2], h5filename, cc[2], cc[1], cc[0], group)
                    geometry[slices] = geo.format(*sig)
                    topology[slices] = get_topology(N, kind=1)
                grid[slices] = ''

        # if slice of data, need to know along which axes
        # Since there may be many different slices, we need to create
        # one xdmf-file for each composition of slices
        attrs = {}
        for tstep in timesteps:
            d = dsets[tstep]
            slx = set()
            for i, x in enumerate(d):
                slices = x.split("/")[2]
                if not 'slice' in slices:
                    slices = 'whole'
                N = NN[slices]
                if slices not in attrs:
                    attrs[slices] = ''
                attrs[slices] += get_attribute(x, h5filename, N, prec)
                slx.add(slices)
            for slices in slx:
                grid[slices] += get_grid(geometry[slices], topology[slices],
                                         attrs[slices].rstrip())
                attrs[slices] = ''
        for slices, ff in xff.items():
            if 'slice' in slices:
                fname = h5filename[:-3]+"_"+slices+".xdmf"
            else:
                fname = h5filename[:-3]+".xdmf"
            xfl = open(fname, "w")
            h = ff.format(tt, len(timesteps), grid[slices].rstrip())
            xfl.write(h)
            xfl.close()
    f.close()

if __name__ == "__main__":
    import sys
    generate_xdmf(sys.argv[-1])



import importlib
import functools
from sympy.matrices.common import NonInvertibleMatrixError
import sympy as sp
import numpy as np

n = sp.Symbol('n', real=True, integer=True, positive=True)

def get_stencil_matrix(bcs, family, alpha=None, beta=None, gn=1):
    """Return stencil matrix.

    Return the narrowest possible stencil matrix satisfying the
    homogeneous boundary conditions in `bcs`, and with leading
    coefficient equal to one.

    Parameters
    ----------
    bcs : dict
        The boundary conditions in dictionary form, see
        :class:`.BoundaryConditions`.

    family : str
        Choose one of

        - ``Chebyshev``
        - ``Chebyshevu``
        - ``Legendre``
        - ``Ultraspherical``
        - ``Jacobi``
        - ``Laguerre``

    alpha, beta : numbers, optional
        The Jacobi parameters, used only for ``Jacobi`` or ``Ultraspherical``

    gn : Scaling function for Jacobi polynomials

    """
    from shenfun.spectralbase import BoundaryConditions
    base = importlib.import_module('.'.join(('shenfun', family.lower())))
    bnd_values = functools.partial(base.Orthogonal.bnd_values, alpha=alpha, beta=beta, gn=gn)
    bcs = BoundaryConditions(bcs)
    bc = {'D': 0, 'N': 1, 'N2': 2, 'N3': 3, 'N4': 4}
    lr = {'L': 0, 'R': 1}
    lra = {'L': 'left', 'R': 'right'}
    s = []
    r = []
    for key in bcs.orderednames():
        k, v = key[0], key[1:]
        if v == 'R':
            alfa = bcs[lra[k]]['R'][0]
            f = [bnd_values(k=0)[lr[k]], bnd_values(k=1)[lr[k]]]
            s.append([sp.simplify(f[0](n+j)+alfa*f[1](n+j)) for j in range(1, 1+bcs.num_bcs())])
            r.append(-sp.simplify(f[0](n)+alfa*f[1](n)))
        else:
            f = bnd_values(k=bc[v])[lr[k]]
            s.append([sp.simplify(f(n+j)) for j in range(1, 1+bcs.num_bcs())])
            r.append(-sp.simplify(f(n)))
    A = sp.Matrix(s)
    b = sp.Matrix(r)
    M = sp.simplify(A.solve(b))
    d = {0: 1}
    for i, s in enumerate(M):
        if not s == 0:
            d[i+1] = s
    return d

def get_bc_basis(bcs, family, alpha=None, beta=None, gn=1):
    """Return boundary basis satisfying `bcs`.

    Parameters
    ----------
    bcs : dict
        The boundary conditions in dictionary form, see
        :class:`.BoundaryConditions`.

    family : str
        Choose one of

        - ``Chebyshev``
        - ``Chebyshevu``
        - ``Legendre``
        - ``Ultraspherical``
        - ``Jacobi``
        - ``Laguerre``

    alpha, beta : numbers, optional
        The Jacobi parameters, used only for ``Jacobi`` or ``Ultraspherical``

    gn : Scaling function for Jacobi polynomials

    """
    from shenfun.spectralbase import BoundaryConditions
    base = importlib.import_module('.'.join(('shenfun', family.lower())))
    bnd_values = functools.partial(base.Orthogonal.bnd_values, alpha=alpha, beta=beta, gn=gn)
    bcs = BoundaryConditions(bcs)
    def _computematrix(first):
        bc = {'D': 0, 'N': 1, 'N2': 2, 'N3': 3, 'N4': 4}
        lr = {'L': 0, 'R': 1}
        lra = {'L': 'left', 'R': 'right'}
        s = []
        for key in bcs.orderednames():
            k, v = key[0], key[1:]
            if v == 'R':
                alfa = bcs[lra[k]]['R'][0]
                f = [bnd_values(k=0)[lr[k]], bnd_values(k=1)[lr[k]]]
                s.append([sp.simplify(f[0](j)+alfa*f[1](j)) for j in range(first, first+bcs.num_bcs())])
            else:
                f = bnd_values(k=bc[v])[lr[k]]
                s.append([sp.simplify(f(j)) for j in range(first, first+bcs.num_bcs())])

        A = sp.Matrix(s)
        s = sp.simplify(A.solve(sp.eye(bcs.num_bcs())).T)
        return s

    first_basis = bcs.num_derivatives() // bcs.num_bcs()
    first = 0
    for first in range(first_basis+1):
        try:
            s = _computematrix(first)
            break
        except NonInvertibleMatrixError:
            continue

    sol = sp.Matrix(np.zeros((bcs.num_bcs(), first+bcs.num_bcs())))
    sol[:, first:] = s
    return sol



#!/usr/bin/env python

import os
import re
import subprocess
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from numpy import get_include

cwd = os.path.abspath(os.path.dirname(__file__))
cdir = os.path.join(cwd, "shenfun", "optimization", "cython")

def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    devnull = open(os.devnull, "w")
    p = subprocess.Popen([compiler.compiler[0], '-E', '-'] + [flagname],
                         stdin=subprocess.PIPE, stdout=devnull, stderr=devnull)
    p.communicate("")
    return True if p.returncode == 0 else False

class build_ext_subclass(build_ext):
    def build_extensions(self):
        extra_compile_args = ['-g0']
        if os.environ.get("READTHEDOCS", None) == "True":
            extra_compile_args.append('-O0')
        else:
            extra_compile_args.append('-O3')
            if not os.environ.get('CONDA_BUILD', 0):
                for c in ['-w', '-Ofast', '-ffast-math', '-march=native']:
                    if has_flag(self.compiler, c):
                        extra_compile_args.append(c)

        for e in self.extensions:
            e.extra_compile_args += extra_compile_args
            e.include_dirs.extend([get_include()])
        build_ext.build_extensions(self)

def get_extensions():
    ext = []
    for s in ("Matvec", "la", "evaluate", "transforms", "leg2chebm"):
        ext.append(Extension("shenfun.optimization.cython.{0}".format(s),
                             libraries=['m'],
                             sources=[os.path.join(cdir, '{0}.pyx'.format(s))],
                             language="c++"))  # , define_macros=define_macros
    ext.append(Extension("shenfun.legendre.fastgl.fastgl_wrap",
                         libraries=['m'],
                         language='c++',
                         sources=[os.path.join(cwd, "shenfun", "legendre", "fastgl", "fastgl_wrap.pyx")]))
    [e.extra_link_args.extend(["-std=c++11"]) for e in ext]
    #[e.extra_link_args.extend(["-std=c++11", "-fopenmp"]) for e in ext]
    for s in ("Cheb", "convolve", "outer", "applymask", "cross"):
        ext.append(Extension("shenfun.optimization.cython.{0}".format(s),
                             libraries=['m'],
                             sources=[os.path.join(cdir, '{0}.pyx'.format(s))]))

    return ext

def version():
    srcdir = os.path.join(cwd, 'shenfun')
    with open(os.path.join(srcdir, '__init__.py')) as f:
        m = re.search(r"__version__\s*=\s*'(.*)'", f.read())
        return m.groups()[0]

with open("README.rst", "r") as fh:
    long_description = fh.read()

if __name__ == '__main__':
    setup(name="shenfun",
          version=version(),
          description="Shenfun -- Automated Spectral-Galerkin framework",
          long_description=long_description,
          author="Mikael Mortensen",
          author_email="mikaem@math.uio.no",
          url='https://github.com/spectralDNS/shenfun',
          classifiers=[
              'Development Status :: 4 - Beta',
              'Environment :: Console',
              'Intended Audience :: Developers',
              'Intended Audience :: Science/Research',
              'Intended Audience :: Education',
              'Programming Language :: Python',
              'Programming Language :: Python :: 3',
              'Programming Language :: Cython',
              'License :: OSI Approved :: BSD License',
              'Topic :: Scientific/Engineering :: Mathematics',
              'Topic :: Software Development :: Libraries :: Python Modules',
              ],
          packages=["shenfun",
                    "shenfun.optimization",
                    "shenfun.optimization.cython",
                    "shenfun.optimization.numba",
                    "shenfun.legendre",
                    "shenfun.legendre.fastgl",
                    "shenfun.laguerre",
                    "shenfun.hermite",
                    "shenfun.chebyshev",
                    "shenfun.chebyshevu",
                    "shenfun.ultraspherical",
                    "shenfun.fourier",
                    "shenfun.jacobi",
                    "shenfun.forms",
                    "shenfun.utilities",
                    "shenfun.io"
                   ],
          package_dir={"shenfun": "shenfun"},
          install_requires=["mpi4py-fft", "mpi4py", "cython", "numpy", "scipy"],
          setup_requires=["numpy>=1.9",
                          "cython>=0.25",
                          "setuptools>=18.0"],
          ext_modules=get_extensions(),
          cmdclass={"build_ext": build_ext_subclass}
         )



import sympy as sp
import numpy as np
from numpy.polynomial import laguerre as lag
from scipy.special import eval_laguerre
from shenfun.matrixbase import SparseMatrix
from shenfun.jacobi.recursions import n
from shenfun.spectralbase import SpectralBase, getCompositeBase, getBCGeneric, BoundaryConditions

#pylint: disable=method-hidden,no-else-return,not-callable,abstract-method,no-member,cyclic-import

bases = ['Orthogonal',
         'CompactDirichlet',
         'CompactNeumann',
         'Generic']
bcbases = ['BCGeneric']
testbases = []
__all__ = bases + bcbases

xp = sp.Symbol('x', real=True)


class Orthogonal(SpectralBase):
    r"""Function space for a regular Laguerre series

    The orthogonal basis is the Laguerre function

    .. math::

        \phi_k = La_k \exp(-x/2), \quad k = 0, 1, \ldots, N-1,

    where :math:`La_k` is the :math:`k`'th Laguerre polynomial.

    Parameters
    ----------
    N : int
        Number of quadrature points
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    Note
    ----
    We are using Laguerre functions and not the regular Laguerre polynomials
    as basis functions.
    """
    def __init__(self, N, dtype=float, padding_factor=1, dealias_direct=False,
                 coordinates=None, **kw):
        SpectralBase.__init__(self, N, quad="LG", domain=(0, sp.S.Infinity), dtype=dtype,
                              padding_factor=padding_factor, dealias_direct=dealias_direct,
                              coordinates=coordinates)
        self.plan(int(N*padding_factor), 0, dtype, {})

    @staticmethod
    def family():
        return 'laguerre'

    def reference_domain(self):
        return (0, sp.S.Infinity)

    def domain_factor(self):
        return 1

    def points_and_weights(self, N=None, map_true_domain=False, weighted=True, **kw):
        if N is None:
            N = self.shape(False)
        if self.quad == "LG":
            points, weights = lag.laggauss(N)
            if weighted:
                weights *= np.exp(points)
        else:
            raise NotImplementedError

        return points, weights

    def vandermonde(self, x):
        V = lag.lagvander(x, int(self.N*self.padding_factor)-1)
        V *= np.exp(-x/2)[:, None]
        return V

    def evaluate_basis(self, x, i=0, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape)
        output_array = eval_laguerre(i, x, out=output_array)
        output_array *= np.exp(-x/2)
        return output_array

    def evaluate_basis_derivative(self, x=None, i=0, k=0, output_array=None):
        if x is None:
            x = self.mesh(False, False)
        if output_array is None:
            output_array = np.zeros(x.shape)
        x = np.atleast_1d(x)
        basis = np.zeros(self.shape(True))
        basis[i] = 1
        basis = lag.Laguerre(basis)
        if k == 1:
            basis = basis.deriv(k)-0.5*basis
        elif k == 2:
            basis = basis.deriv(2)-basis.deriv(1)+0.25*basis
        elif k == 3:
            basis = basis.deriv(3)-1.5*basis.deriv(2)+0.75*basis.deriv(1)-basis/8
        elif k == 4:
            basis = basis.deriv(4)-2*basis.deriv(3)+1.5*basis.deriv(2)-0.5*basis.deriv(1)+basis/16
        output_array[:] = basis(x)*np.exp(-x/2)
        return output_array

    @staticmethod
    def bnd_values(k=0, **kw):
        if k == 0:
            return (lambda i: 1, None)
        elif k == 1:
            return (lambda i: -i-sp.S.Half, None)
        raise NotImplementedError

    def stencil_matrix(self, N=None):
        N = self.N if N is None else N
        return SparseMatrix({0: 1}, (N, N))

    def evaluate_basis_derivative_all(self, x=None, k=0, argument=0):
        if x is None:
            x = self.mesh(False, False)
        V = lag.lagvander(x, int(self.N*self.padding_factor)-1)
        M = V.shape[1]
        if k == 1:
            D = np.zeros((M, M))
            D[:-1, :] = lag.lagder(np.eye(M), 1)
            W = np.dot(V, D)
            W -= 0.5*V
            V = W*np.exp(-x/2)[:, np.newaxis]

        elif k == 2:
            D = np.zeros((M, M))
            D[:-2, :] = lag.lagder(np.eye(M), 2)
            D[:-1, :] -= lag.lagder(np.eye(M), 1)
            W = np.dot(V, D)
            W += 0.25*V
            V = W*np.exp(-x/2)[:, np.newaxis]

        elif k == 3:
            D = np.zeros((M, M))
            D[:-3, :] = lag.lagder(np.eye(M), 3)
            D[:-2, :] -= lag.lagder(np.eye(M), 2)*3/2
            D[:-1, :] += lag.lagder(np.eye(M), 1)*3/4
            W = np.dot(V, D)
            W += V/8
            V = W*np.exp(-x/2)[:, np.newaxis]

        elif k == 4:
            D = np.zeros((M, M))
            D[:-4, :] = lag.lagder(np.eye(M), 4)
            D[:-3, :] -= lag.lagder(np.eye(M), 3)*2
            D[:-2, :] += lag.lagder(np.eye(M), 2)*3/2
            D[:-1, :] -= lag.lagder(np.eye(M), 1)/2
            W = np.dot(V, D)
            W += V/16
            V = W*np.exp(-x/2)[:, np.newaxis]

        elif k == 0:
            V *= np.exp(-x/2)[:, np.newaxis]

        else:
            raise NotImplementedError

        return V

    def eval(self, x, u, output_array=None):
        x = np.atleast_1d(x)
        if output_array is None:
            output_array = np.zeros(x.shape, dtype=self.dtype)
        output_array[:] = lag.lagval(x, u)*np.exp(-x/2)
        return output_array

    def orthogonal_basis_function(self, i=0, x=sp.symbols('x')):
        return sp.laguerre(i, x)*sp.exp(-x/2)

    def L2_norm_sq(self, i):
        return 1

    def l2_norm_sq(self, i=None):
        if i is None:
            return np.ones(self.N)
        return 1

    @property
    def is_orthogonal(self):
        return True

    def get_orthogonal(self, **kwargs):
        d = dict(quad=self.quad,
                 domain=self.domain,
                 dtype=self.dtype,
                 padding_factor=self.padding_factor,
                 dealias_direct=self.dealias_direct,
                 coordinates=self.coors.coordinates)
        d.update(kwargs)
        return Orthogonal(self.N, **d)

    @staticmethod
    def short_name():
        return 'La'

    def get_bc_space(self):
        if self._bc_space:
            return self._bc_space
        self._bc_space = BCGeneric(self.N, bc=self.bcs)
        return self._bc_space

CompositeBase = getCompositeBase(Orthogonal)
BCGeneric = getBCGeneric(CompositeBase)

class CompactDirichlet(CompositeBase):
    r"""Laguerre function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::

        \phi_k &= (La_k - La_{k+1})\exp(-x/2), \, k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= L_0\exp(-x/2),

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u(0) &= a

    The last basis function is for boundary condition and only used if a is
    different from 0. In one dimension :math:`\hat{u}_{N-1}=a`.

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Laguerre-Gauss
    bc : 1-tuple of number (a,)
        Boundary value at x=0
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, bc=(0,), dtype=float, padding_factor=1,
                 dealias_direct=False, coordinates=None, **kw):
        CompositeBase.__init__(self, N, quad='LG', dtype=dtype, padding_factor=padding_factor,
                               bc=bc, dealias_direct=dealias_direct, domain=(0, sp.S.Infinity),
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: -1}

    @staticmethod
    def boundary_condition():
        return 'Dirichlet'

    @staticmethod
    def short_name():
        return 'SD'


class CompactNeumann(CompositeBase):
    r"""Laguerre function space for Dirichlet boundary conditions

    The basis :math:`\{\phi_k\}_{k=0}^{N-1}` is

    .. math::
        \phi_k &= (La_k - \frac{2k+1}{2k+3}La_{k+1})\exp(-x/2), \, k=0, 1, \ldots, N-2, \\
        \phi_{N-1} &= La_0\exp(-x/2),

    such that

    .. math::
        u(x) &= \sum_{k=0}^{N-1} \hat{u}_k \phi_k(x), \\
        u'(0) &= a

    Parameters
    ----------
    N : int
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Laguerre-Gauss
    bc : 1-tuple of number (a,)
        Boundary value a = u'(0)
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, bc=(0,), dtype=float, padding_factor=1,
                 dealias_direct=False, coordinates=None, **kw):
        if isinstance(bc, (tuple, list)):
            assert len(bc) == 1
            bc = BoundaryConditions({'left': {'N': bc[0]}, 'right': {}}, domain=(0, np.inf))
        CompositeBase.__init__(self, N, dtype=dtype, quad='LG', padding_factor=padding_factor,
                               bc=bc, dealias_direct=dealias_direct, domain=(0, sp.S.Infinity),
                               coordinates=coordinates)
        self._stencil = {0: 1, 1: -(2*n+1)/(2*n+3)}

    @staticmethod
    def boundary_condition():
        return 'Neumann'

    @staticmethod
    def short_name():
        return 'CN'


class Generic(CompositeBase):
    r"""Function space for Laguerre space with any boundary conditions

    Any combination of Dirichlet and Neumann is possible.

    Parameters
    ----------
    N : int, optional
        Number of quadrature points
    quad : str, optional
        Type of quadrature

        - LG - Laguerre-Gauss

    bc : dict, optional
        The dictionary must have key 'left' (not 'right'), to describe boundary
        conditions on the left boundary. Specify Dirichlet with

            {'left': {'D': a}}

        for some value `a`, that will be neglected in the current
        function. Specify mixed Neumann and Dirichlet as

            {'left': {'D': a, 'N': b}}

        See :class:`~shenfun.spectralbase.BoundaryConditions`.
    domain : 2-tuple of floats, optional
        The computational domain
    dtype : data-type, optional
        Type of input data in real physical space. Will be overloaded when
        basis is part of a :class:`.TensorProductSpace`.
    padding_factor : float, optional
        Factor for padding backward transforms.
    dealias_direct : bool, optional
        Set upper 1/3 of coefficients to zero before backward transform
    coordinates: 2- or 3-tuple (coordinate, position vector (, sympy assumptions)), optional
        Map for curvilinear coordinatesystem, and parameters to :class:`~shenfun.coordinates.Coordinates`

    """
    def __init__(self, N, quad="LG", bc={}, domain=(0, sp.S.Infinity), dtype=float,
                 padding_factor=1, dealias_direct=False, coordinates=None, **kw):
        from shenfun.utilities.findbasis import get_stencil_matrix
        self._stencil = get_stencil_matrix(bc, 'laguerre')
        if not isinstance(bc, BoundaryConditions):
            bc = BoundaryConditions(bc, domain=domain)
        CompositeBase.__init__(self, N, quad=quad, domain=domain, dtype=dtype, bc=bc,
                               padding_factor=padding_factor, dealias_direct=dealias_direct,
                               coordinates=coordinates)

    @staticmethod
    def boundary_condition():
        return 'Generic'

    @staticmethod
    def short_name():
        return 'GL'



cimport fastgl_wrap
# distutils: language = c++
#cython: boundscheck=False
#cython: wraparound=False
#cython: language_level=3

import numpy as np

def getGLPair(int N, int k):
    """Return point and weight k for N-point Gauss-Legendre rule

    Parameters
    ----------
    N : int
        The total number of quadrature points
    k : int
        The k'th point in the N-point quadrature rule
    """
    f = fastgl_wrap.GLPair(N, k)
    return tuple((f.x(), f.weight))

cdef GLPairs(int N, double [::1] x, double [::1] w):
    cdef:
        int i

    for i in range(N):
        f = fastgl_wrap.GLPair(N, N-i)
        x[i] = f.x()
        w[i] = f.weight

def leggauss(int N):
    """Gauss-Legendre quadrature

    Computes the points and weights for N-point Gauss-Legendre quadrature.

    Parameters
    ----------
    N : int
        The total number of quadrature points/weights

    Note
    ----
    These points and weights are accurate for more than millions of points.
    The code is based on the paper

        "I. Bogaert, 'Iteration-Free Computation of Gauss-Legendre Quadrature
        Nodes and Weights', SIAM Journal on Scientific Computing, 36, 3,
        A1008-A1026, 2014"

    """
    x = np.zeros(N)
    w = np.zeros(N)
    GLPairs(N, x, w)
    return x, w



cimport numpy as np
#cython: boundscheck=False
#cython: wraparound=False
#cython: language_level=3

ctypedef np.complex128_t complex_t
ctypedef np.int64_t int_t

def convolve_1D(np.ndarray[complex_t, ndim=1] u, np.ndarray[complex_t, ndim=1] v,
                np.ndarray[complex_t, ndim=1] uv, np.ndarray[int_t, ndim=1] k):
    cdef int m, n, p, i, j, N
    cdef complex_t um, vn

    N = u.shape[0]
    for m in k:
        for n in k:
            p = m + n
            if N % 2 == 0:
                if abs(m) == N//2:
                    um = u[m]*0.5
                else:
                    um = u[m]
                if abs(n) == N//2:
                    vn = v[n]*0.5
                else:
                    vn = v[n]
            else:
                um = u[m]
                vn = v[n]
            uv[p] = uv[p] + um*vn



def convolve_real_1D(np.ndarray[complex_t, ndim=1] u, np.ndarray[complex_t, ndim=1] v,
                     np.ndarray[complex_t, ndim=1] uv, np.ndarray[int_t, ndim=1] k):
    cdef int m, n, p, i, j, N
    cdef complex_t um, vn

    N = uv.shape[0]-1
    for m in k:
        for n in k:
            p = m + n
            if p >= 0:
                if N % 2 == 0:
                    if abs(m) == N//2:
                        um = u[abs(m)]*0.5
                    elif m >= 0:
                        um = u[m]
                    else:
                        um = u[abs(m)].real - 1j*u[abs(m)].imag
                    if abs(n) == N//2:
                        vn = v[abs(n)]*0.5
                    elif n >= 0:
                        vn = v[n]
                    else:
                        vn = v[abs(n)].real - 1j*v[abs(n)].imag
                else:
                    if m >= 0:
                        um = u[m]
                    elif m < 0:
                        um = u[abs(m)].real - 1j*u[abs(m)].imag
                    if n >= 0:
                        vn = v[n]
                    elif n < 0:
                        vn = v[abs(n)].real - 1j*v[abs(n)].imag
                uv[p] = uv[p] + um*vn



import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD

__all__ = ['LagrangianParticles']

class LagrangianParticles:
    """Class for tracking Lagrangian particles

    Parameters
    ----------
    points : array
        Initial location of particles. (D, N) array, with N particles in D dimensions
    dt : float
        Time step
    u_hat : :class:`.Function`
        Spectral Galerkin :class:`.Function` for the Eulerian velocity

    """

    def __init__(self, points, dt, u_hat):
        self.x = points
        self.u_hat = u_hat
        self.dt = dt
        self.up = np.zeros(self.x.shape)

    def step(self):
        up = self.rhs()
        self.x[:] = self.x + self.dt*up

    def rhs(self):
        return self.u_hat.eval(self.x, output_array=self.up)

if __name__ == '__main__':
    from shenfun import *
    import sympy as sp
    import matplotlib.pyplot as plt
    import h5py

    N = (40, 40)
    # Should work for any of these bases
    #F0 = FunctionSpace(N[0], 'F', dtype='D', domain=(0., 1.))
    #F1 = FunctionSpace(N[1], 'F', dtype='d', domain=(0., 1.))
    F0 = FunctionSpace(N[0], 'C', bc=(0, 0), domain=(0., 1.))
    F1 = FunctionSpace(N[1], 'C', bc=(0, 0), domain=(0., 1.))
    #F0 = FunctionSpace(N[0], 'C', domain=(0., 1.))
    #F1 = FunctionSpace(N[1], 'C', domain=(0., 1.))

    T = TensorProductSpace(comm, (F0, F1))
    TV = VectorSpace(T)

    x, y = sp.symbols("x,y")
    psi = 1./np.pi*sp.sin(np.pi*x)**2*sp.sin(np.pi*y)**2 # Streamfunction
    ux = -psi.diff(y, 1)
    uy = psi.diff(x, 1)

    uxl = sp.lambdify((x, y), ux, 'numpy')
    uyl = sp.lambdify((x, y), uy, 'numpy')
    X = T.local_mesh(True)
    u = Array(T, buffer=uxl(X[0], X[1]))
    v = Array(T, buffer=uyl(X[0], X[1]))
    uv = Function(TV)
    uv[0] = T.forward(u, uv[0])
    uv[1] = T.forward(v, uv[1])

    # Arrange particles in a circle around (0.5, 0.75) with radius 0.15
    t0 = np.linspace(0, 2*np.pi, 100)[:-1]
    points = np.array([0.5+0.15*np.cos(t0), 0.75+0.15*np.sin(t0)])

    # Create LagrangianParticles instance with given points
    dt = 0.001
    lp = LagrangianParticles(points, dt, uv)

    # Store velocity vectors for later plotting on rank 0
    u.write('velocity.h5', name='u', domain=T.mesh())
    v.write('velocity.h5', name='v', domain=T.mesh())
    if comm.Get_rank() == 0:
        f = h5py.File('velocity.h5', 'r+')
        f.create_group('points')

    # Run simulation from time = 0 to 1 forwards, and then integrate back to 0
    end_time = 2.0
    t = 0
    lg = ['Velocity field']
    b = 'Fwd'
    nsteps = int(end_time/dt)+1
    count = 0
    for i in range(nsteps):
        if np.any(np.round(t, 4) in (0, 0.5, 1.0)):
            if comm.Get_rank() == 0:
                f['points'].create_dataset(str(count), shape=points.shape, dtype=float)
                f['points/'+str(count)][:] = lp.x
            print('Storing points at time %2.1f'%t)
            lg.append('%s at %2.1f' %(b, t))
            count += 1
        if i == (nsteps-1)//2:
            lp.dt *= -1
            b = 'Bwd'
            print('Integrate backwards')
        t += lp.dt
        lp.step()

    if comm.Get_rank() == 0:
        plt.quiver(f['u/mesh/x0'], f['u/mesh/x1'], f['u/2D/0'].__array__().T, f['v/2D/0'].__array__().T)
        steps = list(f['points'].keys())
        for step in steps:
            plt.scatter(f['points/'+str(step)][0], f['points/'+str(step)][1])
        plt.title('Particles integrated forwards and backwards')
        plt.legend(lg)
        plt.show()



import warnings
from functools import wraps
import numba as nb
import numpy as np
from scipy.special import gammaln

warnings.simplefilter('ignore', category=nb.core.errors.NumbaPerformanceWarning)

__all__ = ['scalar_product', 'evaluate_expansion_all', 'cheb2leg', 'leg2cheb',
           'restricted_product', '_leg2cheb', 'FMMdirect1', 'FMMdirect2',
           'FMMdirect3', 'FMMdirect4']

def NDim(func):
    @wraps(func)
    def wrapped_function(input_array, output_array, axis, *args):
        n = input_array.ndim
        if n == 1:
            func(input_array, output_array, *args)
        elif n == 2:
            fun_2D(func, input_array, output_array, axis, *args)
        elif n == 3:
            fun_3D(func, input_array, output_array, axis, *args)
        elif n == 4:
            fun_4D(func, input_array, output_array, axis, *args)
    return wrapped_function

@nb.jit(nopython=True, fastmath=True, cache=False)
def _matvec(A, x, b, m, n, transpose):
    if transpose == 0:
        for i in range(m):
            s = 0.0
            a = A[i]
            for j in range(n):
                s = s+a[j]*x[j]
            b[i] = s
    else:
        for j in range(n):
            b[j] = 0
        for i in range(m):
            s = x[i]
            a = A[i]
            for j in range(n):
                b[j] = b[j] + s*a[j]
    return b

@nb.jit(nopython=True, fastmath=True, cache=False)
def fun_2D(fun, input_array, output_array, axis, *args):
    if axis == 0:
        for j in range(input_array.shape[1]):
            fun(input_array[:, j], output_array[:, j], *args)
    elif axis == 1:
        for i in range(input_array.shape[0]):
            fun(input_array[i], output_array[i], *args)

@nb.jit(nopython=True, fastmath=True, cache=False)
def fun_3D(fun, input_array, output_array, axis, *args):
    if axis == 0:
        for j in range(input_array.shape[1]):
            for k in range(input_array.shape[2]):
                fun(input_array[:, j, k], output_array[:, j, k], *args)
    elif axis == 1:
        for i in range(input_array.shape[0]):
            for k in range(input_array.shape[2]):
                fun(input_array[i, :, k], output_array[i, :, k], *args)
    elif axis == 2:
        for i in range(input_array.shape[0]):
            for j in range(input_array.shape[1]):
                fun(input_array[i, j], output_array[i, j], *args)

@nb.jit(nopython=True, fastmath=True, cache=False)
def fun_4D(fun, input_array, output_array, axis, *args):
    if axis == 0:
        for j in range(input_array.shape[1]):
            for k in range(input_array.shape[2]):
                for l in range(input_array.shape[3]):
                    fun(input_array[:, j, k, l], output_array[:, j, k, l], *args)
    elif axis == 1:
        for i in range(input_array.shape[0]):
            for k in range(input_array.shape[2]):
                for l in range(input_array.shape[3]):
                    fun(input_array[i, :, k, l], output_array[i, :, k, l], *args)
    elif axis == 2:
        for i in range(input_array.shape[0]):
            for j in range(input_array.shape[1]):
                for l in range(input_array.shape[3]):
                    fun(input_array[i, j, :, l], output_array[i, j, :, l], *args)
    elif axis == 3:
        for i in range(input_array.shape[0]):
            for j in range(input_array.shape[1]):
                for k in range(input_array.shape[2]):
                    fun(input_array[i, j, k, :], output_array[i, j, k, :], *args)

def scalar_product(input_array, output_array, x, w, axis, a):
    n = input_array.ndim
    if n == 1:
        _scalar_product(input_array, output_array, x, w, a)
    elif n == 2:
        fun_2D(_scalar_product, input_array, output_array, axis, x, w, a)
    elif n == 3:
        fun_3D(_scalar_product, input_array, output_array, axis, x, w, a)
    elif n == 4:
        fun_4D(_scalar_product, input_array, output_array, axis, x, w, a)
    else:
        if axis > 0:
            input_array = np.moveaxis(input_array, axis, 0)
            output_array = np.moveaxis(output_array, axis, 0)
        _scalar_product(input_array, output_array, x, w, a)
        if axis > 0:
            input_array = np.moveaxis(input_array, 0, axis)
            output_array = np.moveaxis(output_array, 0, axis)

def evaluate_expansion_all(input_array, output_array, x, axis, a):
    n = input_array.ndim
    if n == 1:
        _evaluate_expansion_all(input_array, output_array, x, a)
    elif n == 2:
        fun_2D(_evaluate_expansion_all, input_array, output_array, axis, x, a)
    elif n == 3:
        fun_3D(_evaluate_expansion_all, input_array, output_array, axis, x, a)
    elif n == 4:
        fun_4D(_evaluate_expansion_all, input_array, output_array, axis, x, a)
    else:
        if axis > 0:
            input_array = np.moveaxis(input_array, axis, 0)
            output_array = np.moveaxis(output_array, axis, 0)
        _evaluate_expansion_all(input_array, output_array, x, a)
        if axis > 0:
            input_array = np.moveaxis(input_array, 0, axis)
            output_array = np.moveaxis(output_array, 0, axis)

def restricted_product(space, input_array, output_array, x, i0, i1, a0, axis, a):
    n = input_array.ndim
    Lnm = space.evaluate_basis(x[i0:i1], i=a0)
    Ln = space.evaluate_basis(x[i0:i1], i=a0+1)
    if n == 1:
        _restricted_product(input_array, output_array, x, Lnm, Ln, i0, i1, a0, a)
    elif n == 2:
        fun_2D(_restricted_product, input_array, output_array, axis, x, Lnm, Ln, i0, i1, a0, a)
    elif n == 3:
        fun_3D(_restricted_product, input_array, output_array, axis, x, Lnm, Ln, i0, i1, a0, a)
    elif n == 4:
        fun_4D(_restricted_product, input_array, output_array, axis, x, Lnm, Ln, i0, i1, a0, a)
    else:
        if axis > 0:
            input_array = np.moveaxis(input_array, axis, 0)
            output_array = np.moveaxis(output_array, axis, 0)
        _restricted_product(input_array, output_array, x, Lnm, Ln, i0, i1, a0, a)
        if axis > 0:
            input_array = np.moveaxis(input_array, 0, axis)
            output_array = np.moveaxis(output_array, 0, axis)
    return output_array

@nb.jit(nopython=True, fastmath=True, cache=False)
def _scalar_product(input_array, output_array, xj, wj, a):
    M = output_array.shape[0]
    N = xj.shape[0]
    Lnm = np.ones(N)
    if a.shape[0] == 3:
        anm = a[0] # a_{n-1, n}
        ann = a[1] # a_{n, n}
        anp = a[2] # a_{n+1, n}
    else:
        anm = a[0]
        anp = a[1]
        ann = np.zeros(N+2)
    Ln = (xj-ann[0])/anm[0]
    Lnp = (xj-ann[1])/anm[1]*Ln - anp[1]/anm[1]*Lnm
    for k in range(M):
        s1 = 1/anm[k+2]
        s2 = anp[k+2]/anm[k+2]
        a00 = ann[k+2]
        s = 0.0
        for j in range(N):
            s += Lnm[j]*wj[j]*input_array[j]
            Lnm[j] = Ln[j]
            Ln[j] = Lnp[j]
            Lnp[j] = s1*(xj[j]-a00)*Ln[j] - s2*Lnm[j]
        output_array[k] = s

@nb.jit(nopython=True, fastmath=True, cache=False)
def _evaluate_expansion_all(input_array, output_array, xj, a):
    M = input_array.shape[0]
    N = output_array.shape[0]
    Lnm = np.ones(N)
    if a.shape[0] == 3:
        anm = a[0] # a_{n-1, n}
        ann = a[1] # a_{n, n}
        anp = a[2] # a_{n+1, n}
    else:
        anm = a[0]
        anp = a[1]
        ann = np.zeros(M+2)
    Ln = (xj-ann[0])/anm[0]
    Lnp = (xj-ann[1])/anm[1]*Ln - anp[1]/anm[1]*Lnm
    output_array[:] = 0
    for k in range(M):
        s1 = 1/anm[k+2]
        s2 = anp[k+2]/anm[k+2]
        a00 = ann[k+2]
        for j in range(N):
            output_array[j] += Lnm[j]*input_array[k]
            Lnm[j] = Ln[j]
            Ln[j] = Lnp[j]
            Lnp[j] = s1*(xj[j]-a00)*Ln[j] - s2*Lnm[j]

@nb.jit(nopython=True, fastmath=True, cache=False)
def _restricted_product(input_array, output_array, xj, Lnm0, Ln0, i0, i1, a0, a):
    N = xj.shape[0]
    Lnm = Lnm0.copy()
    Ln = Ln0.copy()
    if a.shape[0] == 3:
        anm = a[0] # a_{n-1, n}
        ann = a[1] # a_{n, n}
        anp = a[2] # a_{n+1, n}
    else:
        anm = a[0]
        anp = a[1]
        ann = np.zeros(N+2)
    Lnp = (xj[i0:i1]-ann[1+a0])/anm[1+a0]*Ln - anp[1+a0]/anm[1+a0]*Lnm
    for k in range(len(output_array)):
        kp = k+a0
        s1 = 1/anm[kp+2]
        s2 = anp[kp+2]/anm[kp+2]
        a00 = ann[kp+2]
        s = 0.0
        for j in range(len(Ln)):
            s += Lnm[j]*input_array[i0+j]
            Lnm[j] = Ln[j]
            Ln[j] = Lnp[j]
            Lnp[j] = s1*(xj[i0+j]-a00)*Ln[j] - s2*Lnm[j]
        output_array[k] = s

Omega = lambda z: np.exp(gammaln(z+0.5) - gammaln(z+1))

@nb.jit(nopython=True, fastmath=True, cache=False)
def _leg2cheb(c, v, a, N, transpose):
    M_2_PI = 2/np.pi
    if transpose is False:
        for n in range(0, N, 2):
            #v[:(N-n)] += a[n//2]*a[n//2:(N-n//2)]*c[n:]
            nh = n//2
            for i in range(N-n):
                v[i] += a[nh]*a[nh+i]*c[n+i]
        v[0] /= 2
        v *= M_2_PI
    else:
        cv = c*M_2_PI
        cv[0] /= 2
        for n in range(0, N, 2):
            nh = n//2
            for i in range(N-n):
                v[i+n] += a[nh]*a[nh+i]*cv[i]
            #v[n:] += (a[n//2]*M_2_PI)*a[n//2:(N-n//2)]*c[:(N-n)]

def leg2cheb(input_array, output_array=None, axis=0, transpose=False):
    if output_array is None:
        output_array = np.zeros_like(input_array)
    else:
        output_array.fill(0)
    n = input_array.ndim
    N = input_array.shape[axis]
    k = np.arange(N)
    a = Omega(k)
    if n == 1:
        _leg2cheb(input_array, output_array, a, N, transpose)
    elif n == 2:
        fun_2D(_leg2cheb, input_array, output_array, axis, a, N, transpose)
    elif n == 3:
        fun_3D(_leg2cheb, input_array, output_array, axis, a, N, transpose)
    elif n == 4:
        fun_4D(_leg2cheb, input_array, output_array, axis, a, N, transpose)
    else:
        if axis > 0:
            input_array = np.moveaxis(input_array, axis, 0)
            output_array = np.moveaxis(output_array, axis, 0)
        _leg2cheb(input_array, output_array, a, N, transpose)
        if axis > 0:
            input_array = np.moveaxis(input_array, 0, axis)
            output_array = np.moveaxis(output_array, 0, axis)
    return output_array

@nb.jit(nopython=True, fastmath=True, cache=False)
def _cheb2leg(v, c, a, dn, N):
    vn = v.copy()
    for i in range(1, N):
        vn[i] = v[i]*i
    c[:] = np.sqrt(np.pi)*a*vn
    for n in range(2, N, 2):
        c[:(N-n)] -= dn[n//2]*a[n//2:(N-n//2)]*vn[n:]
    for i in range(N):
        c[i] *= (i+0.5)

def cheb2leg(input_array, output_array=None, axis=0):
    if output_array is None:
        output_array = np.zeros_like(input_array)
    else:
        output_array.fill(0)
    n = input_array.ndim
    N = input_array.shape[axis]
    k = np.arange(N)
    k[0] = 1
    dn = Omega((k[::2]-2)/2)/k[::2]
    a = 1/(2*Omega(k)*k*(k+0.5))
    a[0] = 2/np.sqrt(np.pi)
    if n == 1:
        _cheb2leg(input_array, output_array, a, dn, N)
    elif n == 2:
        fun_2D(_cheb2leg, input_array, output_array, axis, a, dn, N)
    elif n == 3:
        fun_3D(_cheb2leg, input_array, output_array, axis, a, dn, N)
    elif n == 4:
        fun_4D(_cheb2leg, input_array, output_array, axis, a, dn, N)
    else:
        if axis > 0:
            input_array = np.moveaxis(input_array, axis, 0)
            output_array = np.moveaxis(output_array, axis, 0)
        _cheb2leg(input_array, output_array, a, dn, N)
        if axis > 0:
            input_array = np.moveaxis(input_array, 0, axis)
            output_array = np.moveaxis(output_array, 0, axis)
    return output_array


@NDim
@nb.jit(nopython=True, fastmath=True, cache=True)
def FMMdirect1(u, v, a, n0, trans):
    N = u.shape[0]
    if trans is False:
        for n in range(0, n0):
            for i in range(0, N-2*n):
                v[i] += a[n]*a[n+i]*u[2*n+i]
    else:
        for n in range(0, n0):
            for i in range(N-2*n):
                v[i+2*n] += a[n]*a[n+i]*u[i]

@NDim
@nb.jit(nopython=True, fastmath=True, cache=True)
def FMMdirect2(u, v, a, h, Nd, n0, trans):
    N = v.shape[0]
    for k in range(Nd):
        i0 = k*h
        j0 = n0+i0
        for n in range(0, h, 2):
            a0 = i0+(n0+n)//2
            if trans is True:
                for j in range(min(N-(j0+n), h-n)):
                    v[j0+n+j] += a[(n+n0)//2]*a[a0+j]*u[i0+j]
            else:
                for j in range(min(N-(j0+n), h-n)):
                    v[i0+j] += a[(n+n0)//2]*a[a0+j]*u[j0+n+j]

@NDim
@nb.jit(nopython=True, fastmath=True, cache=True)
def FMMdirect3(u, v, dn, a, h, Nd, n0):
    N = v.shape[0]
    for k in range(Nd):
        i0 = k*h
        j0 = n0+i0
        for n in range(0, h, 2):
            a0 = i0+(n0+n)//2
            for j in range(min(N-(j0+n), h-n)):
                v[i0+j] -= dn[(n+n0)//2]*a[a0+j]*u[j0+n+j]

@NDim
@nb.jit(nopython=True, fastmath=True, cache=True)
def FMMdirect4(u, v, dn, a, n0):
    N = u.shape[0]
    v[:] += np.sqrt(np.pi)*a*u
    for n in range(1, n0):
        for i in range(0, N-2*n):
            v[i] -= dn[n]*a[n+i]*u[2*n+i]



import numba as nb
import numpy as np
from .la import Solve_axis_2D, Solve_axis_3D, Solve_axis_4D

__all__ = ['FDMA_LU', 'FDMA_Solve', 'FDMA_inner_solve']

@nb.jit(nopython=True, fastmath=True, cache=True)
def FDMA_LU(data):
    ld = data[0, :-2]
    d = data[1, :]
    u1 = data[2, 2:]
    u2 = data[3, 4:]
    n = int(d.shape[0])
    for i in range(2, n):
        ld[i-2] = ld[i-2]/d[i-2]
        d[i] -= ld[i-2]*u1[i-2]
        if i < n-2:
            u1[i] -= ld[i-2]*u2[i-2]

def FDMA_Solve(x, data, axis=0):
    n = x.ndim
    if n == 1:
        FDMA_inner_solve(x, data)
    elif n == 2:
        Solve_axis_2D(data, x, FDMA_inner_solve, axis)
    elif n == 3:
        Solve_axis_3D(data, x, FDMA_inner_solve, axis)
    elif n == 4:
        Solve_axis_4D(data, x, FDMA_inner_solve, axis)
    else:
        if axis > 0:
            x = np.moveaxis(x, axis, 0)
        FDMA_inner_solve(x, data)
        if axis > 0:
            x = np.moveaxis(x, 0, axis)

@nb.njit
def FDMA_inner_solve(u, data):
    ld = data[0, :-2]
    d = data[1, :]
    u1 = data[2, 2:]
    u2 = data[3, 4:]
    n = d.shape[0]
    for i in range(2, n):
        u[i] -= ld[i-2]*u[i-2]
    u[n-1] = u[n-1]/d[n-1]
    u[n-2] = u[n-2]/d[n-2]
    u[n-3] = (u[n-3] - u1[n-3]*u[n-1])/d[n-3]
    u[n-4] = (u[n-4] - u1[n-4]*u[n-2])/d[n-4]
    for i in range(n - 5, -1, -1):
        u[i] = (u[i] - u1[i]*u[i+2] - u2[i]*u[i+4])/d[i]



import os
from collections.abc import Mapping
import yaml

# The configuration can be overloaded by a local 'shenfun.yaml' file, or
# in '~/.shenfun/shenfun.yaml'. A yaml file to work with can be created
# using the `dumpconfig` function below

config = {
    'optimization':
    {
        'mode': 'cython',
        'verbose': False,
    },
    'basisvectors': 'normal',
    'transforms':
    {
        'kind':
        {
            'chebyshev': 'fast',
            'chebyshevu': 'fast',
            'fourier': 'fast',
            'legendre': 'recursive',
            'ultraspherical': 'recursive',
            'hermite': 'vandermonde',
            'laguerre': 'vandermonde',
            'jacobi': 'recursive'
        }
    },
    'matrix':
    {
        'sparse':
        {
            #'permc_spec': 'NATURAL',
            'permc_spec': 'COLAMD',
            'solve': 'csc',
            'diags': 'csc',
            'matvec': 'csr'
        },
        'block':
        {
            'assemble': 'csc',
            'use_scipy': True,
            'permc_spec': 'COLAMD',
        }
    },
    'bases':
    {
        'legendre':
        {
            'mode': 'numpy',
            #'mode': 'mpmath'
            'precision': 30,
        },
        'jacobi':
        {
            'mode': 'numpy',
            #'mode': 'mpmath',
            'precision': 30
        }
    },
    'fftw':
    {
        'dct':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE',
        },
        'dst':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE',
        },
        'rfft':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE'
        },
        'irfft':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE'
        },
        'fft':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE'
        },
        'ifft':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE'
        },
        'dlt':
        {
            'threads': 1,
            'planner_effort': 'FFTW_MEASURE'
        }
    }
}

def update(conf, newconf):
    """Recursive update"""
    for key, value in newconf.items():
        if isinstance(value, Mapping) and value:
            returned = update(conf.get(key, {}), value)
            conf[key] = returned
        else:
            conf[key] = newconf[key]
    return conf

locations = [os.path.expanduser('~/.shenfun'),
             os.getcwd()]

for loc in locations:
    fl = os.path.expandvars(os.path.join(loc, 'shenfun.yaml'))
    try:
        with open(fl, 'r') as yf:
            update(config, yaml.load(yf, Loader=yaml.FullLoader))
        yf.close()
    except FileNotFoundError:
        pass

def dumpconfig(filename='shenfun.yaml', path='~/.shenfun'): # pragma: no cover
    """Dump a configuration file in yaml format
    """
    with open(os.path.join(os.path.expanduser(path), filename), 'w') as yf:
        yaml.dump(config, yf)
    yf.close()
